int inputX;
int inputY;
int inputLength;
int inputType;
int inputNewNumber;
int inputReturnMode;
bool inputClearOnFirstChange;
int inputCursor;
char inputNewString[MAX_TEXT_WDT];
int* inputOriginalNumber;
char* inputOriginalString;
bool inputNumberSigned;
bool inputDone;

#define INPUT_TYPE_STRING		0
#define INPUT_TYPE_NUMBER		1
#define INPUT_TYPE_HEX_STRING	2



void set_input_number(int x,int y,int len,int mode,bool changec,int* number,bool sign)
{
	memset(inputNewString,32,sizeof(inputNewString));

	inputNewNumber=*number;
	inputOriginalNumber=number;

	inputType=INPUT_TYPE_NUMBER;
	inputX=x;
	inputY=y;
	inputLength=len;
	inputReturnMode=mode;
	inputClearOnFirstChange=changec;
	inputNumberSigned=sign;

	set_edit_mode(MODE_INPUT);
}



int input_actual_string_length(void)
{
	int i;

	for(i=inputLength-1;i>=0;--i)
	{
		if(inputNewString[i]!=' ') break;
	}

	return i;
}



void set_input_string(int x,int y,int length,int mode,bool changec,char* string)
{
	memset(inputNewString,32,sizeof(inputNewString));
	memcpy(inputNewString,string,strlen(string));

	inputOriginalString=string;

	inputType=INPUT_TYPE_STRING;
	inputX=x;
	inputY=y;
	inputLength=length;
	inputReturnMode=mode;
	inputClearOnFirstChange=changec;
	inputCursor=input_actual_string_length()+1;

	set_edit_mode(MODE_INPUT);
}



void set_input_hex_string(int x,int y,int length,int mode,char* string)
{
	memset(inputNewString,0,sizeof(inputNewString));
	memcpy(inputNewString,string,length);

	inputOriginalString=string;

	inputType=INPUT_TYPE_HEX_STRING;
	inputX=x;
	inputY=y;
	inputLength=length;
	inputReturnMode=mode;
	inputCursor=0;

	set_edit_mode(MODE_INPUT);
}



void draw_input(void)
{
	int i;

	switch(inputType)
	{
	case INPUT_TYPE_NUMBER:
		{
			if(inputNumberSigned) PUT_CHAR(inputX-1,inputY,(inputNewNumber<0)?'-':' ');

			put_num_attr(inputX,inputY,abs(inputNewNumber),inputLength,COL_CURSOR);
		}
		break;

	case INPUT_TYPE_STRING:
		{
			for(i=0;i<inputLength;++i)
			{
				PUT_CHAR_ATTR(inputX+i,inputY,inputNewString[i],(i==inputCursor)?COL_CURSOR:COL_CURSOR_ROW);
			}
		}
		break;

	case INPUT_TYPE_HEX_STRING:
		{
			for(i=0;i<inputLength;++i)
			{
				PUT_CHAR_ATTR(inputX+i,inputY,NUM2HEX(inputNewString[i]),(i==inputCursor)?COL_CURSOR:COL_CURSOR_ROW);
			}
		}
		break;
	}
}



void input_events(SDL_Event event)
{
	int key;
	bool sign;

	switch(event.type)
	{
	case SDL_KEYDOWN:
		{
			key=event.key.keysym.sym;

			switch(inputType)
			{
			case INPUT_TYPE_NUMBER:
				{
					switch(key)
					{
					case SDLK_ESCAPE:
						{
							inputDone=false;

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_RETURN:
						{
							inputDone=true;

							*inputOriginalNumber=inputNewNumber;

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_BACKSPACE:
						{
							inputNewNumber/=10;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_DELETE:
						{
							inputNewNumber=0;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_MINUS:
						{
							if(inputNumberSigned)
							{
								inputNewNumber=0-inputNewNumber;
							}
						}
						break;
					}

					if(key>=SDLK_0&&key<=SDLK_9)
					{
						if(inputClearOnFirstChange)
						{
							inputNewNumber=0;
							inputClearOnFirstChange=false;
						}

						sign=inputNewNumber<0?true:false;

						inputNewNumber=(abs(inputNewNumber)*10+(key-SDLK_0))%(int)pow(10.0f,inputLength);

						if(sign) inputNewNumber=0-inputNewNumber;
					}
				}
				break;

			case INPUT_TYPE_STRING:
				{
					switch(key)
					{
					case SDLK_ESCAPE:
						{
							inputDone=false;

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_RETURN:
						{
							inputDone=true;

							inputNewString[input_actual_string_length()+1]=0;
							strcpy(inputOriginalString,inputNewString);

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_BACKSPACE:
						{
							if(inputCursor)
							{
								memcpy(&inputNewString[inputCursor-1],&inputNewString[inputCursor],inputLength-1);

								--inputCursor;
							}

							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_LEFT:
						{
							if(inputCursor) --inputCursor;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_RIGHT:
						{
							if(inputCursor<=input_actual_string_length()) ++inputCursor;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_HOME:
						{
							inputCursor=0;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_END:
						{
							inputCursor=input_actual_string_length()+1;
							inputClearOnFirstChange=false;
						}
						break;

					case SDLK_DELETE:
						{
							memcpy(&inputNewString[inputCursor],&inputNewString[inputCursor+1],inputLength-inputCursor-1);
							inputClearOnFirstChange=false;
						}
						break;
					}

					key = event.key.keysym.sym;

					if (key >= 32 && key < 128)
					{
						if (SDL_GetModState() & KMOD_SHIFT)
						{
							if (key >= 'a'&&key <= 'z')  key -= 0x20;
							if (key == '1') key = '!';
							if (key == '2') key = '@';
							if (key == '3') key = '#';
							if (key == '4') key = '$';
							if (key == '5') key = '%';
							if (key == '6') key = '^';
							if (key == '7') key = '&';
							if (key == '8') key = '*';
							if (key == '9') key = '(';
							if (key == '0') key = ')';
							if (key == '-') key = '_';
							if (key == '=') key = '+';
							if (key == '\\') key = '|';
							if (key == '[') key = '{';
							if (key == ']') key = '}';
							if (key == ';') key = ':';
							if (key == '\'') key = '\"';
							if (key == ',') key = '<';
							if (key == '.') key = '>';
							if (key == '/') key = '?';
							if (key == '`') key = '~';
						}

						if(inputClearOnFirstChange)
						{
							memset(inputNewString,32,sizeof(inputNewString));
							inputClearOnFirstChange=false;
							inputCursor=0;
						}

						if(input_actual_string_length()<inputLength-2)
						{
							memcpy(&inputNewString[inputCursor+1],&inputNewString[inputCursor],sizeof(inputNewString)-inputCursor-1);

							inputNewString[inputCursor++]=key;

							if(inputCursor>inputLength-1) inputCursor=inputLength-1;
						}
					}
				}
				break;

			case INPUT_TYPE_HEX_STRING:
				{
					switch(key)
					{
					case SDLK_ESCAPE:
						{
							inputDone=false;

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_RETURN:
						{
							inputDone=true;

							memcpy(inputOriginalString,inputNewString,inputLength);

							set_edit_mode(inputReturnMode);
						}
						break;

					case SDLK_BACKSPACE:
						{
							inputNewString[inputCursor]=0;

							if(inputCursor) --inputCursor;
						}
						break;

					case SDLK_LEFT:
						{
							if(inputCursor) --inputCursor;
						}
						break;

					case SDLK_RIGHT:
						{
							if(inputCursor<inputLength-1) ++inputCursor;
						}
						break;

					case SDLK_HOME:
						{
							inputCursor=0;
						}
						break;

					case SDLK_END:
						{
							inputCursor=inputLength-1;
						}
						break;

					case SDLK_DELETE:
						{
							inputNewString[inputCursor]=0;

							if(inputCursor<inputLength-1) ++inputCursor;
						}
						break;
					}

					if(key>=SDLK_0&&key<=SDLK_9)
					{
						inputNewString[inputCursor]=key-SDLK_0;

						if(inputCursor<inputLength-1) ++inputCursor;
					}

					if(key>=SDLK_a&&key<=SDLK_f)
					{
						inputNewString[inputCursor]=key-SDLK_a+10;

						if(inputCursor<inputLength-1) ++inputCursor;
					}
				}
				break;
			}
		}
		break;
	}
}