#define MAX_INSTRUMENT_GUI	64

enum {
	INSGUI_SIGNED_BYTE = 0,
	INSGUI_UNSIGNED_BYTE,
	INSGUI_SIGNED_WORD,
	INSGUI_UNSIGNED_WORD,
	INSGUI_HEX_STRING,
	INSGUI_OPTION,
	INSGUI_LABEL,
	INSGUI_ENVELOPE,
	INSGUI_SELECT_FILE
};

struct instrumentGuiStruct {
	int type;
	int min;
	int max;
	int digits;
	int byte;
	int length;
	int option_count;
	int y;
	int res;

	char caption[128];
	char option[128];
	char fileext[8];
};

struct instrumentGuiStruct instrumentGuiList[MAX_INSTRUMENT_GUI];

int instrumentGuiCount;
int instrumentGuiCursor=0;
int instrumentGuiNumber;
int instrumentGuiParamXOff;
int instrumentGuiActiveControls;

bool instrumentGuiEnvelope;
int instrumentGuiEnvelopeCursor;



void instrument_gui_add_number(int type,const char* caption,int min,int max,int byte)
{
	int i,digits;
	
	i=instrumentGuiCount;

	instrumentGuiList[i].type=type;
	instrumentGuiList[i].min=min;
	instrumentGuiList[i].max=max;

	if(min>max) max=min;

	digits=1;

	if(max>=10)    digits=2;
	if(max>=100)   digits=3;
	if(max>=1000)  digits=4;
	if(max>=10000) digits=5;

	instrumentGuiList[i].digits=digits;
	instrumentGuiList[i].byte=byte;

	strcpy(instrumentGuiList[i].caption,caption);

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void instrument_gui_add_hex_string(const char* caption,int length,int max,int res,int byte,bool env)
{
	int i;

	i=instrumentGuiCount;

	instrumentGuiList[i].type = !env ? INSGUI_HEX_STRING : INSGUI_ENVELOPE;
	instrumentGuiList[i].length = length;
	instrumentGuiList[i].byte = byte;
	instrumentGuiList[i].max = max;
	instrumentGuiList[i].res = res;

	strcpy(instrumentGuiList[i].caption,caption);

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void instrument_gui_add_option(const char* caption,const char* option,int byte)
{
	unsigned int i,n;

	n=1;

	for(i=0;i<strlen(option);++i) if(option[i]=='|') ++n;

	i=instrumentGuiCount;

	instrumentGuiList[i].type=INSGUI_OPTION;
	instrumentGuiList[i].byte=byte;
	instrumentGuiList[i].option_count=n;

	strcpy(instrumentGuiList[i].option,option);
	strcpy(instrumentGuiList[i].caption,caption);

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void instrument_gui_add_label(const char* caption)
{
	unsigned int i;

	i=instrumentGuiCount;

	instrumentGuiList[i].type=INSGUI_LABEL;

	strcpy(instrumentGuiList[i].caption,caption);

	++instrumentGuiCount;
}



void instrument_gui_add_select_file(const char* caption, const char* fileext)
{
	unsigned int i;

	i = instrumentGuiCount;

	instrumentGuiList[i].type = INSGUI_SELECT_FILE;

	strcpy(instrumentGuiList[i].caption, caption);
	strcpy(instrumentGuiList[i].fileext, fileext);

	++instrumentGuiCount;
	++instrumentGuiActiveControls;
}



void set_instrument_editor(void)
{
	int i, y, len;

	memset(instrumentGuiList, 0, sizeof(instrumentGuiList));

	instrumentGuiCount = 0;
	++instrumentGuiActiveControls = 0;
	instrumentGuiEnvelope = false;
	instrumentGuiEnvelopeCursor = 0;

	change_instrument();

	len = 0;
	y = 0;

	for (i = 0; i < instrumentGuiCount; ++i)
	{
		instrumentGuiList[i].y = y;

		if (instrumentGuiList[i].type != INSGUI_ENVELOPE)
		{
			++y;
		}
		else
		{
			y += instrumentGuiList[i].max / instrumentGuiList[i].res;
		}

		if (instrumentGuiList[i].type == INSGUI_LABEL) continue;

		if ((int)strlen(instrumentGuiList[i].caption) > len)
		{
			len = strlen(instrumentGuiList[i].caption);
		}
	}

	if (instrumentGuiCount)
	{
		while (instrumentGuiList[instrumentGuiCursor].type == INSGUI_LABEL&&instrumentGuiCursor < instrumentGuiCount) ++instrumentGuiCursor;

		if (instrumentGuiCursor >= instrumentGuiCount) instrumentGuiCursor = instrumentGuiCount - 1;
	}

	instrumentGuiParamXOff=PATTERN_XPOS+1+len+2;

	set_edit_mode(MODE_INSTRUMENT_EDITOR);
}



int instrument_get_value(int i)
{
	unsigned char* data;
	int n, min, max;

	data=module.instrument[curInstrument].data+instrumentGuiList[i].byte;

	switch(instrumentGuiList[i].type)
	{
	case INSGUI_SIGNED_BYTE:   n = *((char*)data); break;
	case INSGUI_UNSIGNED_BYTE: n = *((unsigned char*)data); break;
	case INSGUI_SIGNED_WORD:   n = *((short*)data); break;
	case INSGUI_UNSIGNED_WORD: n = *((unsigned short*)data); break;
	default: n = 0;
	}

	min = instrumentGuiList[i].min;
	max = instrumentGuiList[i].max;

	if (n < min) n = min;
	if (n > max) n = max;

	return n;
}



void instrument_set_value(int i,int n)
{
	int min,max;
	unsigned char* data;

	data=module.instrument[curInstrument].data+instrumentGuiList[i].byte;

	min=instrumentGuiList[i].min;
	max=instrumentGuiList[i].max;

	if(n<min) n=min;
	if(n>max) n=max;

	switch(instrumentGuiList[i].type)
	{
	case INSGUI_SIGNED_BYTE:   *(char*)data=(char)n; break;
	case INSGUI_UNSIGNED_BYTE: *(unsigned char*)data=(unsigned char)n; break;
	case INSGUI_SIGNED_WORD:   *(short*)data=(short)n; break;
	case INSGUI_UNSIGNED_WORD: *(unsigned short*)data=(unsigned short)n; break;
	}
}



void instrument_add_value(int value)
{
	int i;
	unsigned char* data;

	i=instrumentGuiCursor;

	data=module.instrument[curInstrument].data+instrumentGuiList[i].byte;

	switch(instrumentGuiList[i].type)
	{
	case INSGUI_SIGNED_BYTE:
	case INSGUI_UNSIGNED_BYTE:
	case INSGUI_SIGNED_WORD:
	case INSGUI_UNSIGNED_WORD:
		{
			instrument_set_value(i,instrument_get_value(i)+value);
		}
		break;

	case INSGUI_OPTION:
		{
			if(value<0)
			{
				if(*data>0) --*data; else *data=instrumentGuiList[i].option_count-1;
			}
			else
			{
				if(*data<instrumentGuiList[i].option_count-1) ++*data; else *data=0;
			}
		}
		break; 
	}

	set_instrument_editor();
}



void draw_instrument_editor(void)
{
	char str[1024],c;
	int i,j,k,x1,x2,y,d,min,max,val;
	unsigned char* data;
	unsigned char col;
	char* option;

	for(i=0;i<textScreenWidth;++i)
	{
		PUT_ATTR(i,0,COL_HEADER);
		PUT_ATTR(i,textScreenHeight-1,COL_HEADER);
	}

	put_str(1,0,"Instrument editor",COL_HEADER);

	if(!module.instruments)
	{
		put_str(PATTERN_XPOS+1,PATTERN_YPOS+1,"This engine does not support instruments",COL_PROGRAM_BACKGROUND);

		return;
	}

	put_str(1,textScreenHeight-1,"Instrument     Channel --  ~F2~Save~F3~Load~F4~Back~F5~PlayRow",COL_HEADER);
	put_num(12,textScreenHeight-1,curInstrument+1,3);
	if(module.columnChannel[curColumn]>=0) put_num(24,textScreenHeight-1,module.columnChannel[curColumn],2);

	x1=PATTERN_XPOS+1;
	x2=instrumentGuiParamXOff+10;

	for(i=0;i<instrumentGuiCount;++i)
	{
		y=PATTERN_YPOS+1+instrumentGuiList[i].y;

		col = (i == instrumentGuiCursor) ? COL_CURSOR_ROW : COL_PROGRAM_BACKGROUND;

		if (instrumentGuiList[i].type == INSGUI_LABEL) col = COL_PROGRAM_BACKGROUND;

		put_str(x1,y,instrumentGuiList[i].caption,col);
		put_str(x1 - 1, y, " ", col);
		put_str(x1 + strlen(instrumentGuiList[i].caption), y, " ", col);

		d=instrumentGuiList[i].digits;

		data=&module.instrument[curInstrument].data[instrumentGuiList[i].byte];
		
		min=instrumentGuiList[i].min;
		max=instrumentGuiList[i].max;

		switch(instrumentGuiList[i].type)
		{
		case INSGUI_SIGNED_BYTE:
		case INSGUI_SIGNED_WORD:
			{
				if (instrument_get_value(i)<0) PUT_CHAR(instrumentGuiParamXOff - 1, y, '-');
			}
		case INSGUI_UNSIGNED_BYTE:
		case INSGUI_UNSIGNED_WORD:
			{
				put_num(instrumentGuiParamXOff, y, abs(instrument_get_value(i)), d);

				sprintf(str,"(%i..%i)",min,max);

				put_str(x2, y, str, COL_PROGRAM_BACKGROUND);
			}
			break;

		case INSGUI_HEX_STRING:
			{
				for(j=0;j<instrumentGuiList[i].length;++j)
				{
					PUT_CHAR(instrumentGuiParamXOff+j,y,NUM2HEX(data[j]));
				}
			}
			break;

		case INSGUI_OPTION:
			{
				option=instrumentGuiList[i].option;

				k=*data;

				if(k)
				{
					for(j=0;j<(int)strlen(option);++j)
					{
						if(option[j]=='|')
						{
							--k;

							if(!k)
							{
								option=&option[j+1];
								break;
							}
						}
					}
				}

				for(j=0;j<(int)strlen(option);++j)
				{
					c=option[j];

					if(c=='|') break;

					PUT_CHAR(instrumentGuiParamXOff+j,y,c);
				}
			}
			break;

		case INSGUI_ENVELOPE:
			{
				for(j=0;j<instrumentGuiList[i].length;++j)
				{
					val=module.instrument[curInstrument].data[instrumentGuiList[i].byte+j];

					if (instrumentGuiList[i].res == 1)
					{
						for (k = 0; k < instrumentGuiList[i].max; ++k)
						{
							c = 4;

							if (k <= val) c = 5;

							PUT_CHAR(instrumentGuiParamXOff + j, y + (instrumentGuiList[i].max - 1 - k), c);
							PUT_ATTR(instrumentGuiParamXOff + j, y + (instrumentGuiList[i].max - 1 - k), (instrumentGuiEnvelope&&j == instrumentGuiEnvelopeCursor) ? COL_CURSOR : COL_PROGRAM_BACKGROUND);
						}
					}

					if (instrumentGuiList[i].res == 2)
					{
						for (k = 0; k < instrumentGuiList[i].max; k += 2)
						{
							c = 0;

							if (k <= val) c = 1;
							if (k < val)  c = 2;

							PUT_CHAR(instrumentGuiParamXOff + j, y + (instrumentGuiList[i].max - 1 - k) / 2, c);
							PUT_ATTR(instrumentGuiParamXOff + j, y + (instrumentGuiList[i].max - 1 - k) / 2, (instrumentGuiEnvelope&&j == instrumentGuiEnvelopeCursor) ? COL_CURSOR : COL_PROGRAM_BACKGROUND);
						}
					}
				}
			}
			break;
		}

		++y;
	}
}



void instrument_editor_events(SDL_Event event)
{
	bool sign;
	int i,val,off;

	if(editMode==MODE_INSTRUMENT_INPUT_DONE)
	{
		instrument_set_value(instrumentGuiCursor,instrumentGuiNumber);

		set_edit_mode(MODE_INSTRUMENT_EDITOR);

		set_instrument_editor();
	}

	switch(event.type)
	{
	case SDL_KEYDOWN:
		{
			if(!instrumentGuiEnvelope)
			{
				//instrument editor input

				//first buttons that is always active

				switch (event.key.keysym.sym)
				{
				case SDLK_ESCAPE:
				case SDLK_F4:
				{
					set_edit_mode(MODE_EDIT_SONG);
				}
				return;

				case SDLK_F5:
				{
					music_play_row();
				}
				break;

				case SDLK_EQUALS:
				case SDLK_KP_PLUS:
				{
					if (curInstrument<module.instruments - 1) ++curInstrument;

					set_instrument_editor();
				}
				break;

				case SDLK_MINUS:
				case SDLK_KP_MINUS:
				{
					if (curInstrument) --curInstrument;

					set_instrument_editor();
				}
				break;

				case SDLK_TAB://move cursor to the next or previous channel
				{
					if (event.key.keysym.mod&(KMOD_LSHIFT | KMOD_RSHIFT)) change_channel(-1); else change_channel(1);

					set_instrument_editor();
				}
				break;
				}

				//then buttons that only active when there is instruments support and active controls in the list

				if (!module.instruments || !instrumentGuiActiveControls) return;

				switch(event.key.keysym.sym)
				{
				case SDLK_UP:
					{
						while(1)
						{
							--instrumentGuiCursor;

							if(instrumentGuiCursor<0) instrumentGuiCursor=instrumentGuiCount-1;

							if(instrumentGuiList[instrumentGuiCursor].type!=INSGUI_LABEL) break;
						}
					}
					break;

				case SDLK_DOWN:
					{
						while(1)
						{
							++instrumentGuiCursor;

							if(instrumentGuiCursor>=instrumentGuiCount) instrumentGuiCursor=0;

							if(instrumentGuiList[instrumentGuiCursor].type!=INSGUI_LABEL) break;
						}
					}
					break;

				case SDLK_PAGEUP:
					{
						instrumentGuiCursor = 0;

						while (instrumentGuiList[instrumentGuiCursor].type == INSGUI_LABEL) ++instrumentGuiCursor;
					}
					break;

				case SDLK_PAGEDOWN:
					{
						instrumentGuiCursor = instrumentGuiCount - 1;

						while (instrumentGuiList[instrumentGuiCursor].type == INSGUI_LABEL) --instrumentGuiCursor;
					}
					break;

				case SDLK_F2:
					{
						set_select_file("Save instrument file", instrumentDirectory, MODE_INSTRUMENT_EDITOR, INSTRUMENT_EXTENSION, true, EVENT_SAVE_INSTRUMENT, selectFileInstrumentLast);
					}
					break;

				case SDLK_F3:
					{
						set_select_file("Load instrument file", instrumentDirectory, MODE_INSTRUMENT_EDITOR, INSTRUMENT_EXTENSION, true, EVENT_LOAD_INSTRUMENT, selectFileInstrumentLast);
					}
					break;

				case SDLK_LEFT:
					{
						instrument_add_value(-1);
					}
					break;

				case SDLK_RIGHT:
					{
						instrument_add_value(1);
					}
					break;

				case SDLK_RETURN:
					{
						switch(instrumentGuiList[instrumentGuiCursor].type)
						{
							case INSGUI_HEX_STRING:
							{
								set_input_hex_string(instrumentGuiParamXOff,instrumentGuiList[instrumentGuiCursor].y+PATTERN_YPOS+1,instrumentGuiList[instrumentGuiCursor].length,MODE_INSTRUMENT_INPUT_DONE,(char*)&module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte]);
							}
							break;

							case INSGUI_ENVELOPE:
							{
								instrumentGuiEnvelope=true;

								if(instrumentGuiEnvelopeCursor>instrumentGuiList[instrumentGuiCursor].length-1) instrumentGuiEnvelopeCursor=instrumentGuiList[instrumentGuiCursor].length-1;
							}
							break;

							case INSGUI_OPTION:
							{
								instrument_add_value(1);
							}
							break;

							case INSGUI_SELECT_FILE:
							{
								set_select_file("Import file", itemSelectFilename, MODE_INSTRUMENT_EDITOR, instrumentGuiList[instrumentGuiCursor].fileext, true, EVENT_SCRIPT_FILE_SELECT, selectFileInstrumentLast);
							}
							break;

							default:
							{
								instrumentGuiNumber=instrument_get_value(instrumentGuiCursor);

								switch(instrumentGuiList[instrumentGuiCursor].type)
								{
								case INSGUI_SIGNED_BYTE:
								case INSGUI_SIGNED_WORD:
									sign=true;
									break;

								default:
									sign=false;
								}

								set_input_number(instrumentGuiParamXOff,instrumentGuiList[instrumentGuiCursor].y+PATTERN_YPOS+1,instrumentGuiList[instrumentGuiCursor].digits,MODE_INSTRUMENT_INPUT_DONE,true,&instrumentGuiNumber,sign);
							}
						}					
					}
					break;
				}
			}
			else
			{
				//envelope editor input

				switch(event.key.keysym.sym)
				{
					case SDLK_ESCAPE:
					case SDLK_RETURN:
					{
						instrumentGuiEnvelope=false;
					}
					break;

					case SDLK_F5:
					{
						music_play_row();
					}
					break;

					case SDLK_UP:
					{
						val=module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte+instrumentGuiEnvelopeCursor];

						++val;

						if (val >= instrumentGuiList[instrumentGuiCursor].max) val = 0;

						module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte+instrumentGuiEnvelopeCursor]=val;
					}
					break;

					case SDLK_DOWN:
					{
						val=module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte+instrumentGuiEnvelopeCursor];

						--val;

						if (val < 0) val = instrumentGuiList[instrumentGuiCursor].max - 1;

						module.instrument[curInstrument].data[instrumentGuiList[instrumentGuiCursor].byte+instrumentGuiEnvelopeCursor]=val;
					}
					break;

					case SDLK_LEFT:
					{
						--instrumentGuiEnvelopeCursor;

						if(instrumentGuiEnvelopeCursor<0) instrumentGuiEnvelopeCursor=instrumentGuiList[instrumentGuiCursor].length-1;
					}
					break;

					case SDLK_RIGHT:
					{
						++instrumentGuiEnvelopeCursor;

						if(instrumentGuiEnvelopeCursor>=instrumentGuiList[instrumentGuiCursor].length) instrumentGuiEnvelopeCursor=0;
					}
					break;

					case SDLK_INSERT:
					{
						off=instrumentGuiList[instrumentGuiCursor].byte;

						for(i=instrumentGuiList[instrumentGuiCursor].length-1;i>instrumentGuiEnvelopeCursor;--i)
						{
							module.instrument[curInstrument].data[off+i]=module.instrument[curInstrument].data[off+i-1];
						}

						++instrumentGuiEnvelopeCursor;

						if(instrumentGuiEnvelopeCursor>=instrumentGuiList[instrumentGuiCursor].length) instrumentGuiEnvelopeCursor=instrumentGuiList[instrumentGuiCursor].length;
					}
					break;

					case SDLK_BACKSPACE:
					{
						if(instrumentGuiEnvelopeCursor)
						{
							off=instrumentGuiList[instrumentGuiCursor].byte;

							for(i=instrumentGuiEnvelopeCursor;i<instrumentGuiList[instrumentGuiCursor].length-1;++i)
							{
								module.instrument[curInstrument].data[off+i]=module.instrument[curInstrument].data[off+i+1];
							}

							module.instrument[curInstrument].data[off+instrumentGuiList[instrumentGuiCursor].length-1]=0;

							--instrumentGuiEnvelopeCursor;
						}
					}
					break;

					case SDLK_DELETE:
					{
						off=instrumentGuiList[instrumentGuiCursor].byte;

						for(i=instrumentGuiEnvelopeCursor;i<instrumentGuiList[instrumentGuiCursor].length-1;++i)
						{
							module.instrument[curInstrument].data[off+i]=module.instrument[curInstrument].data[off+i+1];
						}

						module.instrument[curInstrument].data[off+instrumentGuiList[instrumentGuiCursor].length-1]=0;
					}
					break;
				}
			}
		}
		break;
	}
}