void module_clear(void)
{
	int i;

	memset(&module,0,sizeof(moduleStruct));

	curRow=0;
	curColumn=0;
	curSelectionActive=false;
	curInstrument=0;
	curAutoStepCount=0;
	curResetValue=true;

	module.loopStart=0;
	module.loopEnd=MAX_ROWS-1;
	module.speed=6;
	module.speed_min = 1;
	module.speed_max = 99;
	module.measure=4;

	strcpy(module.title ,"untitled");
	strcpy(module.author,"uncredited");

	update_selection();

	for(i=0;i<MAX_COLUMNS;++i) channelMute[i]=false;

	undoAvailable=false;

	change_engine();
}



int module_find_first_empty_row(void)
{
	int i,j,rows;

	rows=0;

	for(i=MAX_ROWS-1;i>=0;--i)
	{
		if(module.marker[i])
		{
			rows=i+1;
			break;
		}

		for(j=0;j<module.columns;++j)
		{
			if(module.columnType[j]!=COLUMN_TYPE_EMPTY)
			{
				if(module.rows[i].column[j])
				{
					rows=i+1;
					i=-1;
					break;
				}
			}
		}
	}

	return rows;
}



int instrument_get_actual_length(int instrument)
{
	int i;

	for(i=MAX_INSTRUMENT_DATA-1;i>=0;--i)
	{
		if(module.instrument[instrument].data[i]) return i+1;
	}

	return 0;
}



bool load_save_error(const char *str, const char *name)
{
	sprintf(engineErrorMessage, "%s '%s'", str, name);

	return false;
}



bool module_save(const char* filename)
{
	FILE *file;
	int i, j, k, ct, val, row, col, rows, len;

	file=fopen(filename,"wt");

	if (!file) return load_save_error("Can't save module", filename);

	//signature

	fprintf(file,"%s\n\n",SONG_SIGNATURE);

	//engine

	fprintf(file,"Engine=%s\n\n",engineFileName);

	//song settings

	fprintf(file,"[Settings]\n\n");

	fprintf(file,"Title=%s\n" ,module.title);
	fprintf(file,"Author=%s\n",module.author);
	fprintf(file,"Speed=%i\n",module.speed);
	fprintf(file,"LoopStart=%i\n",module.loopStart);
	fprintf(file,"LoopEnd=%i\n",module.loopEnd);
	fprintf(file,"Measure=%i\n",module.measure);

	//pattern format, to keep it human readible in the text, and be able to easily parse back

	fprintf(file,"PatternFormat=");

	for(i=0;i<module.columns;++i)
	{
		if(module.columnType[i]==COLUMN_TYPE_EMPTY)
		{
			fprintf(file,"_");
		}
		else
		{
			fprintf(file,"%i",module.columnType[i]);
		}
	}

	fprintf(file,"\n\n");

	//instruments

	fprintf(file,"[Instruments]\n\n");

	for(i=0;i<module.instruments;++i)
	{
		len=instrument_get_actual_length(i);

		if(len)
		{
			fprintf(file,"Instrument%i=",i+1);

			for(j=0;j<len;++j) fprintf(file,"%2.2x",module.instrument[i].data[j]);

			fprintf(file,"\n");
		}
	}

	fprintf(file,"\n");

	//song text

	rows=module_find_first_empty_row();

	fprintf(file,"[Song]\n\n");

	for (row = 0; row<rows; ++row)
	{
		fprintf(file, "%4.4i%c", row, module.marker[row] ? '*' : ' ');

		for (col = 0; col<module.columns; ++col)
		{
			ct = module.columnType[col];
			val = module.rows[row].column[col];

			switch(ct)
			{
			case COLUMN_TYPE_EMPTY:
				fprintf(file," ");
				break;

			case COLUMN_TYPE_NOTE:
				fprintf(file,"%s",noteNames[val]);
				break;

			default:
				for(k=0;k<ct;++k) fprintf(file,"%c",NUM2DEC(val,ct-1-k));
			}
		}

		if (module.markerName[row]) fprintf(file,"%s", module.markerName[row]);

		fprintf(file,"\n");
	}

	fclose(file);

	return true;
}



int module_find_tag(char* data,int size,const char* tag)
{
	int ptr,len;

	len=strlen(tag);
	ptr=0;

	while(size)
	{
		if(!memcmp(&data[ptr],tag,len)) return ptr+len;

		++ptr;
		--size;
	}

	return -1;
}



char* module_load_str(char* data,int size,const char* tag)
{
	static char str[1024];
	char c;
	int ptr,len;

	len=strlen(tag);
	ptr=0;

	while(size)
	{
		if(!memcmp(data,tag,len))
		{
			data+=len;

			while(size)
			{
				c=*data++;
				if(c<32) break;

				str[ptr++]=c;
			}

			str[ptr]=0;

			return str;
		}

		++data;
		--size;
	}

	return NULL;
}



int module_load_int(char* data,int size,const char* tag)
{
	char c;
	int len,n;

	len=strlen(tag);

	while(size)
	{
		if(!memcmp(data,tag,len))
		{
			data+=len;
			n=0;

			while(size)
			{
				c=*data++;
				if(c<'0'||c>'9') break;

				n=n*10+(c-'0');
				--size;
			}

			return n;
		}

		++data;
		--size;
	}

	return -1;
}



bool module_load(const char* filename)
{
	FILE *file;
	char *text;
	char insName[1024];
	int i, j, n, m, ptr, row, ct, size, insSize, nibble;
	char markerName[MAX_MARKER_NAME + 1];

	file=fopen(filename,"rb");

	if(!file) return load_save_error("Can't load module", filename);

	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);

	text=(char*)malloc(size+1);
	text[size-1]=0;

	fread(text,size,1,file);
	fclose(file);

	//check signature

	if(module_find_tag(text,size,SONG_SIGNATURE)<0)
	{
		free(text);
		return false;
	}

	//song settings

	strcpy(engineFileName,module_load_str(text,size,"Engine="));

	change_engine();
	module_clear();

	module.speed    =module_load_int(text,size,"Speed=");
	module.loopStart=module_load_int(text,size,"LoopStart=");
	module.loopEnd  =module_load_int(text,size,"LoopEnd=");
	module.measure  =module_load_int(text,size,"Measure=");

	strcpy(module.title ,module_load_str(text,size,"Title="));
	strcpy(module.author,module_load_str(text,size,"Author="));

	//instruments

	for(i=0;i<module.instruments;++i)
	{
		sprintf(insName,"Instrument%i=",i+1);

		ptr=module_find_tag(text,size,insName);

		if(ptr>=0)
		{
			insSize=0;
			nibble=0;

			while(ptr<size)
			{
				n=text[ptr++];

				if(n<32) break;

				if(!nibble) m=HEX2DEC(n)<<4; else module.instrument[i].data[insSize++]=m|HEX2DEC(n);

				nibble^=1;
			}
		}
	}

	//song text

	ptr=module_find_tag(text,size,"[Song]");

	while(ptr<size)
	{
		if(text[ptr]>32) break;
		++ptr;
	}

	row=0;

	while(ptr<size)
	{
		while(text[ptr]>='0'&&text[ptr]<='9') ++ptr;

		if(text[ptr++]=='*') module.marker[row]=true;

		for(i=0;i<module.columns;++i)
		{
			ct=module.columnType[i];

			switch(ct)
			{
			case COLUMN_TYPE_EMPTY:
				{
					++ptr;
				}
				break;

			case COLUMN_TYPE_NOTE:
				{
					for(j=0;j<noteNamesCount;++j)
					{
						if(!memcmp(&text[ptr],noteNames[j],3))
						{
							module.rows[row].column[i]=j;
							break;
						}
					}

					ptr+=3;
				}
				break;

			default:
				{
					n=0;

					for(j=0;j<ct;++j)
					{
						m=text[ptr++];

						if(m=='.') m=0; else m=m-'0';

						n=n*10+m;
					}

					module.rows[row].column[i]=n;
				}
			}
		}

		//read optional marker name

		while (text[ptr]==32) ++ptr;
		
		n = 0;

		while (text[ptr] >= 32)
		{
			if (n < MAX_MARKER_NAME) markerName[n++] = text[ptr];

			++ptr;
		}

		if (n)
		{
			markerName[n] = 0;

			strcpy(module.markerName[row], markerName);
		}

		//skip end of the line

		while(text[ptr]<32) ++ptr;

		++row;
	}

	free(text);

	return true;
}



bool instrument_save(const char* filename,int ins)
{
	FILE *file;
	int len,i;

	file=fopen(filename,"wt");

	if (!file) return load_save_error("Can't save instrument", filename);

	//signature

	fprintf(file,"%s\n\n",INSTRUMENT_SIGNATURE);

	//instruments

	len=instrument_get_actual_length(ins);

	if(len)
	{
		fprintf(file,"InstrumentData=");

		for(i=0;i<len;++i) fprintf(file,"%2.2x",module.instrument[ins].data[i]);

		fprintf(file,"\n");
	}

	fprintf(file,"\n");

	fclose(file);
	
	return true;
}



bool instrument_load(const char* filename,int ins)
{
	FILE *file;
	char *text;
	int n,m,ptr,size,insSize,nibble;

	file=fopen(filename,"rb");

	if(!file) return load_save_error("Can't load instrument", filename);

	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);

	text=(char*)malloc(size+1);
	text[size-1]=0;

	fread(text,size,1,file);
	fclose(file);

	//check signature

	if(module_find_tag(text,size,INSTRUMENT_SIGNATURE)<0)
	{
		free(text);
		return false;
	}

	ptr=module_find_tag(text,size,"InstrumentData=");

	if(ptr>=0)
	{
		memset(module.instrument[ins].data,0,MAX_INSTRUMENT_DATA);

		insSize=0;
		nibble=0;

		while(ptr<size)
		{
			n=text[ptr++];

			if(n<32) break;

			if(!nibble) m=HEX2DEC(n)<<4; else module.instrument[ins].data[insSize++]=m|HEX2DEC(n);

			nibble^=1;
		}
	}
	else
	{
		return false;
	}

	free(text);

	return true;
}



void do_backup(void)
{
	char backupName[MAXPATH];
	time_t rawTime;
	struct tm *timeInfo;

	if (!backupPeriod) return;

	time(&rawTime);
	timeInfo = localtime(&rawTime);

	sprintf(backupName, "%s%2.2i%2.2i%2.2i_%2.2i%2.2i.1tm", backupDirectory, timeInfo->tm_mday, timeInfo->tm_mon, timeInfo->tm_year % 100, timeInfo->tm_hour, timeInfo->tm_min);

	module_save(backupName);
}