#ifdef WIN32
#define BUILD_WIN32	//enable to build win32 version without the console window
#endif

#ifdef WIN32
#include <windows.h>
#include <direct.h>
#include "dirent/dirent.h"
#else
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#endif


#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string>
#include <vector>

using std::string;
using std::vector;

#include <SDL.h>

#include <angelscript.h>
#include <as_jit.h>
#include <scriptarray/scriptarray.h>
#include <scriptstdstring/scriptstdstring.h>
#include <scriptbuilder/scriptbuilder.h>

#include "gme/gme.h"



#ifndef MAXPATH
#define MAXPATH	4096
#endif

#define TITLE					"1tracker v0.29"
#define CONFIG_NAME				"1tracker.cfg"

#define ENGINE_EXTENSION		".1te"
#define SONG_EXTENSION			".1tm"
#define INSTRUMENT_EXTENSION	".1ti"

SDL_Window *screenWindow = NULL;
SDL_Renderer *screenRenderer = NULL;
SDL_Texture *screenTexture = NULL;

unsigned int *screenData = NULL;
unsigned int screenPalette[256];

FILE *logFile=NULL;

bool fullScreen;

char baseDirectory[MAXPATH];
char songDirectory[MAXPATH];
char exportDirectory[MAXPATH];
char instrumentDirectory[MAXPATH];
char backupDirectory[MAXPATH];
char itemSelectFilename[MAXPATH];

unsigned int backupPeriod;
unsigned int backupPrevTime;

unsigned char* fontData=NULL;

#define MAX_TEXT_WDT	256
#define MAX_TEXT_HGT	256

int charWidth;
int charHeight;
int charSize;

int screenWidth;
int screenHeight;

char textScreenFont[MAXPATH];
char textScreenColor[MAXPATH];

int textScreenWidth;
int textScreenHeight;

int asJITEnable;

unsigned char textBuf    [MAX_TEXT_WDT*MAX_TEXT_HGT];
unsigned char textBufPrev[MAX_TEXT_WDT*MAX_TEXT_HGT];
unsigned char attrBuf    [MAX_TEXT_WDT*MAX_TEXT_HGT];
unsigned char attrBufPrev[MAX_TEXT_WDT*MAX_TEXT_HGT];

#define POS(x,y)				(((y)*MAX_TEXT_WDT)+(x))
#define PUT_CHAR(x,y,c)			{ textBuf[POS((x),(y))]=(c); }
#define PUT_ATTR(x,y,a)			{ attrBuf[POS((x),(y))]=(a); }
#define PUT_CHAR_ATTR(x,y,c,a)	{ textBuf[POS((x),(y))]=(c); attrBuf[POS((x),(y))]=(a); }

#define NUM2HEX(num)			(((num)?((num)<10?'0'+(num):'A'-10+(num)):'.'))

#define NUM2DECZ(num,digit)		('0'+(num)/(int)pow(10.0f,(digit))%10)
#define NUM2DEC(num,digit)		((num)?NUM2DECZ((num),(digit)):'.')

#define HEX2DEC(x)				((((x)>='0'&&(x)<='9')?(x)-'0':(x)-'a'+10))

#define MAX_COLUMNS				128
#define MAX_ROWS				10000
#define MAX_INSTRUMENTS			256
#define MAX_INSTRUMENT_DATA		16384
#define MAX_MARKER_NAME			8

#define COLUMN_TYPE_EMPTY			-1
#define COLUMN_TYPE_NOTE			0
#define COLUMN_TYPE_NUMBER(digits)	(digits)

#define PATTERN_HEIGHT			(textScreenHeight-2)
#define PATTERN_XPOS			0
#define PATTERN_YPOS			1

#define ENGINE_LIST_HEIGHT		(PATTERN_HEIGHT/2+1)

//top nibble is the index in font.bmp for background color, bottom nibble is index for character color

#define COL_PATTERN_BACKGROUND	0x21
#define COL_CURSOR				0x01
#define COL_CURSOR_ROW			0x31
#define COL_SELECTION			0x51
#define COL_SELECTION_BEAT		0x61
#define COL_SELECTION_CURSOR	0x71
#define COL_PATTERN_BEAT		0x41
#define COL_HEADER				0x89
#define COL_HEADER_MUTE			0xa9
#define COL_DIRECTORY			0xfb
#define COL_ERROR_BACKGROUND	0xc1
#define COL_PROGRAM_BACKGROUND	0xf1
#define COL_MARKER_LIST			0xfd

enum {
	EVENT_LOAD_SONG = 0,
	EVENT_SAVE_SONG,
	EVENT_EXPORT_FILE,
	EVENT_LOAD_INSTRUMENT,
	EVENT_SAVE_INSTRUMENT,
	EVENT_SCRIPT_FILE_SELECT
};

int curRow;
int curColumn;
int curOctave;
int curAutoStep;
int curAutoStepCount;
int curSelectionXS;
int curSelectionXE;
int curSelectionYS;
int curSelectionYE;
int curSelectionX;
int curSelectionY;
int curSelectionWidth;
int curSelectionHeight;
bool curSelectionActive;
bool curResetValue;

int curPageStep;
int curPageLargeStep;

int curInstrument;

bool curEnterHold;
bool curShiftHold;

enum {
	MODE_EDIT_SONG = 0,
	MODE_INPUT,
	MODE_SELECT_FILE,
	MODE_SELECT_FILE_INPUT_DONE,
	MODE_SELECT_FILE_OVERWRITE,
	MODE_INSTRUMENT_EDITOR,
	MODE_INSTRUMENT_INPUT_DONE,
	MODE_INFO,
	MODE_SELECT_ENGINE,
	MODE_HELP,
	MODE_EXPORT
};

enum {
	EDIT_STYLE_LINEAR=0,
	EDIT_STYLE_CHOPPED
};

int editMode;
int editStyle;

bool quietMode;

bool altMenu;

char engineFileName[MAXPATH];
char engineTitle[128];
char engineErrorMessage[1024];

asIScriptEngine *script_engine = NULL;
asCJITCompiler* script_engine_jit = NULL;

char engineHeader[MAX_TEXT_WDT];

#define SONG_SIGNATURE			"[1tracker module]"
#define INSTRUMENT_SIGNATURE	"[1tracker instrument]"

struct rowStruct {
	unsigned char column[MAX_COLUMNS];
};

struct instrumentStruct {
	unsigned char data[MAX_INSTRUMENT_DATA];
};

struct moduleStruct {
	rowStruct     rows  [MAX_ROWS];

	unsigned char marker[MAX_ROWS];
	char          markerName[MAX_ROWS][MAX_MARKER_NAME + 1];

	instrumentStruct instrument[MAX_INSTRUMENTS];

	int columns;					//number of columns in module
	int columnType    [MAX_COLUMNS];//column type
	int columnChannel [MAX_COLUMNS];//assignment to channels (-1 for extra columns)
	int columnMaxValue[MAX_COLUMNS];//max value for parameter columns
	bool columnPlay   [MAX_COLUMNS];//play current row at column update

	int lowestNote;
	int highestNote;

	bool loopSupport;
	int loopStart;
	int loopEnd;

	int instruments;

	int speed;
	int speed_min;
	int speed_max;

	int measure;

	char title[128];
	char author[128];
};

struct markerListStruct {
	int row[MAX_ROWS];
	int length;
	int cur;
};



struct moduleStruct module;
struct moduleStruct moduleUndo;

struct markerListStruct markerList;

bool undoAvailable;

struct rowStruct copyPasteData[MAX_ROWS];
int copyPasteType[MAX_COLUMNS];

int copyPasteWidth;
int copyPasteHeight;

bool channelMute[MAX_COLUMNS];


const int noteNamesCount=2+12*10;

const char *noteNames[noteNamesCount]={
"...","R--",//0,1
"C-0","C#0","D-0","D#0","E-0","F-0","F#0","G-0","G#0","A-0","A#0","B-0",//starting from 2
"C-1","C#1","D-1","D#1","E-1","F-1","F#1","G-1","G#1","A-1","A#1","B-1",
"C-2","C#2","D-2","D#2","E-2","F-2","F#2","G-2","G#2","A-2","A#2","B-2",
"C-3","C#3","D-3","D#3","E-3","F-3","F#3","G-3","G#3","A-3","A#3","B-3",
"C-4","C#4","D-4","D#4","E-4","F-4","F#4","G-4","G#4","A-4","A#4","B-4",
"C-5","C#5","D-5","D#5","E-5","F-5","F#5","G-5","G#5","A-5","A#5","B-5",
"C-6","C#6","D-6","D#6","E-6","F-6","F#6","G-6","G#6","A-6","A#6","B-6",
"C-7","C#7","D-7","D#7","E-7","F-7","F#7","G-7","G#7","A-7","A#7","B-7",
"C-8","C#8","D-8","D#8","E-8","F-8","F#8","G-8","G#8","A-8","A#8","B-8",
"C-9","C#9","D-9","D#9","E-9","F-9","F#9","G-9","G#9","A-9","A#9","B-9"
};

const int noteKeys[]={
	SDLK_z,0,
	SDLK_s,1,
	SDLK_x,2,
	SDLK_d,3,
	SDLK_c,4,
	SDLK_v,5,
	SDLK_g,6,
	SDLK_b,7,
	SDLK_h,8,
	SDLK_n,9,
	SDLK_j,10,
	SDLK_m,11,
	SDLK_COMMA,12,
	SDLK_l,13,
	SDLK_PERIOD,14,
	SDLK_SEMICOLON,15,
	SDLK_SLASH,16,
	SDLK_q,12,
	SDLK_2,13,
	SDLK_w,14,
	SDLK_3,15,
	SDLK_e,16,
	SDLK_r,17,
	SDLK_5,18,
	SDLK_t,19,
	SDLK_6,20,
	SDLK_y,21,
	SDLK_7,22,
	SDLK_u,23,
	SDLK_i,24,
	SDLK_9,25,
	SDLK_o,26,
	SDLK_0,27,
	SDLK_p,28,
	SDLK_LEFTBRACKET,29,
	SDLK_EQUALS,30,
	SDLK_RIGHTBRACKET,31,
	SDLK_BACKSLASH,32,
	SDLK_1,-1,
	SDLK_a,-1
};

const int noteKeysCount=sizeof(noteKeys)/sizeof(int);

bool soundRenderEnable;

Music_Emu* emu;

int emu_speed_factor;

unsigned char* scriptGeneratedData;
int scriptGeneratedDataSize;



void log_open(const char *name)
{
	logFile=fopen(name,"wt");
}



void log_add(const char *format,...)
{
	va_list arg;

	if(!logFile) return;

	va_start(arg,format);
#ifndef BUILD_WIN32
	printf(format,arg);
#endif
	vfprintf(logFile,format,arg);
	va_end(arg);
	fflush(logFile);
}



void log_close(void)
{
	if(logFile) fclose(logFile);
}



void sdl_draw_text_screen(unsigned int *screen,bool all)
{
	int i,j,k,l,n,x,y,ptr;
	unsigned char *src;
	unsigned int *dst;
	unsigned char ink,paper;

	if(all)
	{
		memset(textBufPrev,0xff,sizeof(textBufPrev));
		memset(attrBufPrev,0xff,sizeof(attrBufPrev));
	}
	
	y=0;

	for(i=0;i<textScreenHeight;++i)
	{
		ptr=i*MAX_TEXT_WDT;
		x=0;

		for(j=0;j<textScreenWidth;++j)
		{
			n=textBuf[ptr];

			if(n!=textBufPrev[ptr]||attrBuf[ptr]!=attrBufPrev[ptr])
			{
				if(n>=128) n=32;

				textBufPrev[ptr]=n;
				attrBufPrev[ptr]=attrBuf[ptr];

				src=&fontData[n*charSize];

				ink=   attrBuf[ptr]    &15;
				paper=(attrBuf[ptr]>>4)&15;

				for(k=0;k<charHeight;++k)
				{
					dst=(unsigned int*)&screen[(y+k)*screenWidth+x];

					for(l=0;l<charWidth;++l) *dst++=screenPalette[(*src++?ink:paper)];
				}
			}

			++ptr;
			x+=charWidth;
		}

		y+=charHeight;
	}
}



unsigned int read_dword(unsigned char* data)
{
	return data[0]+(data[1]<<8)+(data[2]<<16)+(data[3]<<24);
}



unsigned int read_word(unsigned char* data)
{
	return data[0]+(data[1]<<8);
}



bool load_font(const char* filename)
{
	FILE *file;
	int i, j, k, l, x, y, rle, bpp, size, offset, wdt, hgt, pitch, ps, pd;
	unsigned char *data;
	bool flip;

	file=fopen(filename,"rb");

	if(!file) return false;

	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);

	data=(unsigned char*)malloc(size);

	fread(data,size,1,file);
	fclose(file);
	
	if(data[0]!='B'||data[1]!='M')
	{
		log_add("Error: Not BMP file format (%s)\n",filename);
		return false;
	}
	
	offset=read_dword(&data[10]);
	wdt=read_dword(&data[18]);
	hgt=read_dword(&data[22]);
	bpp=read_word (&data[28]);
	rle=read_dword(&data[30]);
	
	if(rle!=0||bpp!=1)
	{
		log_add("Error: Only uncompressed monochrome BMP supported (%s)\n",filename);
		return false;
	}

	if(hgt>0)
	{
		flip=false;
	}
	else
	{
		hgt=0-hgt;
		flip=true;
	}

	charWidth =wdt/16;
	charHeight=hgt/8;
	charSize  =charWidth*charHeight;

	pitch = ((wdt + 7) / 8 + 3) / 4 * 4;

	fontData = (unsigned char*)malloc(8 * 16 * charSize);

	pd = 0;

	for(i=0;i<8;++i)
	{
		for(j=0;j<16;++j)
		{
			for (k = 0;k < charHeight;++k)
			{
				for (l = 0;l < charWidth;++l)
				{
					x = j*charWidth + l;
					y = i*charHeight + k;

					if (flip)
					{
						ps = offset + y*pitch + x / 8;

					}
					else
					{
						ps = offset + (hgt - 1 - y)*pitch + x / 8;
					}

					fontData[pd++] = (data[ps] & (0x80 >> (x & 7)) ? 1 : 0);
				}
			}
		}
	}

	free(data);

	return true;
}



bool load_color(const char* filename)
{
	FILE *file;
	int i, ptr, size;
	unsigned char *data;

	file = fopen(filename, "rb");

	if (!file) return false;

	fseek(file, 0, SEEK_END);
	size = ftell(file);
	fseek(file, 0, SEEK_SET);

	data = (unsigned char*)malloc(size);

	fread(data, size, 1, file);
	fclose(file);

	if (data[0] != 'B' || data[1] != 'M')
	{
		log_add("Error: Not BMP file format (%s)\n", filename);
		return false;
	}

	if (read_word(&data[28]) != 8)
	{
		log_add("Error: Only 256 colors BMP supported (%s)\n", filename);
		return false;
	}

	ptr = 54;

	for (i = 0;i < 256;++i)
	{
		screenPalette[i] = data[ptr + 0] | (data[ptr + 1] << 8) | (data[ptr + 2] << 16);

		ptr += 4;
	}

	free(data);

	return true;
}



void sdl_flip(void)
{
	SDL_UpdateTexture(screenTexture, NULL, screenData, screenWidth * sizeof(Uint32));
	SDL_RenderClear(screenRenderer);
	SDL_RenderCopy(screenRenderer, screenTexture, NULL, NULL);
	SDL_RenderPresent(screenRenderer);
}



void sdl_set_video(bool update)
{
	if (screenData) free(screenData);

	if (screenTexture) SDL_DestroyTexture(screenTexture);
	if (screenRenderer) SDL_DestroyRenderer(screenRenderer);
	if (screenWindow) SDL_DestroyWindow(screenWindow);

	screenData = (unsigned int*)malloc(screenWidth*screenHeight*sizeof(unsigned int));

	screenWindow = SDL_CreateWindow(TITLE, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, screenWidth, screenHeight, fullScreen ? SDL_WINDOW_FULLSCREEN_DESKTOP : 0);

	if (!screenWindow)
	{
		log_add("Error: Can't create SDL window\n");
		exit(-1);
	}
	
	screenRenderer = SDL_CreateRenderer(screenWindow, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);

	if (!screenRenderer)
	{
		log_add("Error: Can't create SDL renderer\n");
		exit(-1);
	}

	screenTexture = SDL_CreateTexture(screenRenderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STREAMING, screenWidth, screenHeight);

	if (!screenTexture)
	{
		log_add("Error: Can't create SDL texture for viewport\n");
		exit(-1);
	}

	SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "0");
	SDL_ShowCursor(SDL_ENABLE);

	if(update)
	{
		sdl_draw_text_screen(screenData,true);
		sdl_flip();
	}
}



int put_str(int x,int y,const char* str,unsigned char col);

void clear_screen(unsigned char);


void put_str_wrap(const char* str,int xs,int ys,int wdt,int hgt)
{
	int i,x,y,len;
	unsigned int ptr,ptr1;
	char c;

	ptr=0;

	x=xs;
	y=ys;

	while(ptr<strlen(str))
	{
		ptr1=ptr;
		len=0;

		while(ptr1<strlen(str))
		{
			c=str[ptr1];

			if(c==' '||c=='\n') break;

			++len;
			++ptr1;
		}

		if(x+len>=xs+wdt)
		{
			x=xs;
			++y;
		}

		for(i=0;i<len;++i)
		{
			PUT_CHAR(x,y,str[ptr++]);
			++x;
		}

		while(1)
		{
			c=str[ptr];

			if(c=='\n')
			{
				x=xs;
				++y;
				++ptr;
				continue;
			}

			if(c==' ')
			{
				++x;
				++ptr;
				continue;
			}

			break;
		}

		if(y-ys>=hgt) break;
	}
}



void update_screen(void)
{
	sdl_draw_text_screen(screenData, false);
	sdl_flip();

	SDL_Delay(0);
}



void critical_error(const char *format,...)
{
	char str[4096];
	SDL_Event event;
	va_list arg;

	va_start(arg,format);
	log_add(format,arg);
	va_end(arg);

	clear_screen(COL_ERROR_BACKGROUND);

	sprintf(str,format,arg);

	put_str(1,1,"Unrecoverable error",COL_ERROR_BACKGROUND);
	put_str_wrap(str,1,3,textScreenWidth-2,textScreenHeight-4);

	update_screen();

	SDL_Delay(250);

	while(SDL_WaitEvent(&event)>=0)
	{
		switch(event.type)
		{
		case SDL_KEYDOWN:
		case SDL_QUIT:
			exit(0);
			break;
		}

		SDL_Delay(0);
	}

	exit(-1);
}



void handle_gme_error(gme_err_t error,const char *name)
{
	if (error) critical_error("[GME] %s %s\n", name, error);
}



void sdl_sound_callback(void* userdata,unsigned char* stream,int bytes)
{
	short *ptr;

	ptr=(short*)stream;

	if(!soundRenderEnable)
	{
		memset(ptr, 0, bytes);
	}
	else
	{
		handle_gme_error(gme_play(emu, bytes / sizeof(short), ptr), "gme_play");
	}
}



bool compile_song(int format,int start,bool oneshot,bool mute);



void music_stop(void)
{
	if(!soundRenderEnable) return;

	soundRenderEnable=false;

	SDL_LockAudio();
	SDL_UnlockAudio();

	gme_delete(emu);
}



void music_play(int start,bool oneshot)
{
	music_stop();

	//compile and play

	if(compile_song(0,start,oneshot,true))
	{
		if(scriptGeneratedData&&scriptGeneratedDataSize)
		{
			handle_gme_error(gme_open_data(scriptGeneratedData, scriptGeneratedDataSize, &emu, 44100 / emu_speed_factor),"gme_open_data");
			gme_ignore_silence(emu, 1);
			handle_gme_error(gme_start_track(emu, 0),"gme_start_track");

			soundRenderEnable = true;

			free(scriptGeneratedData);
			scriptGeneratedData = NULL;
		}
	}
}



void music_play_from_start(void)
{
	music_play(0,false);
}



void music_play_from_cursor(void)
{
	music_play(curRow,false);
}



void music_play_row(void)
{
	music_play(curRow,true);
}



void music_toggle(bool fromCur)
{
	if(soundRenderEnable)
	{
		music_stop();
	}
	else
	{
		if(fromCur) music_play_from_cursor(); else music_play_from_start();
	}
}



bool music_is_playing(void)
{
	return soundRenderEnable;
}




void update_selection(void);
void change_engine(void);
void change_instrument(void);
void change_channel(int);
bool engine_get_info(char*);
void set_undo();


void set_edit_mode(int mode)
{
	SDL_Event userEvent;

	if(mode!=MODE_EDIT_SONG&&mode!=MODE_INPUT) music_stop();

	editMode=mode;

	userEvent.type=SDL_USEREVENT;
	userEvent.user.code=-1;

	SDL_PushEvent(&userEvent);
}



#include "key_mapping.h"
#include "config.h"
#include "gui.h"
#include "input.h"
#include "select_file.h"
#include "load_save.h"
#include "instrument_editor.h"
#include "info.h"
#include "select_engine.h"
#include "help.h"
#include "export.h"
#include "song_editor.h"
#include "script.h"



void cleanup(void)
{
	free_file_list();

	if(fontData) free(fontData);
	if(script_engine) script_shut();
	
	music_stop();
	log_close();

	SDL_Quit();
}



int main(int argc, char* argv[])
{
	char moduleFileName[MAXPATH];
	char fontFileName  [MAXPATH];
	char colorFileName[MAXPATH];
	char configFileName[MAXPATH];
	char logFileName   [MAXPATH];
	SDL_Event event;
	SDL_AudioSpec *sndReq,*sndObt;
	int i,soundBufferLength;
	bool logEnable;
	FILE *file;

	strcpy(moduleFileName,argc>1?argv[1]:"");

	strcpy(baseDirectory, SDL_GetBasePath());

	i=strlen(baseDirectory);

	while(--i>=0)
	{
		if((baseDirectory[i]=='\\') || (baseDirectory[i]=='/'))
		{
			baseDirectory[i+1]=0;
			break;
		}
	}
	
	strcpy(songDirectory, baseDirectory);
	strcpy(exportDirectory, baseDirectory);
	strcpy(instrumentDirectory, baseDirectory);
	strcpy(itemSelectFilename, baseDirectory);

	strcpy(backupDirectory,baseDirectory);
	strcat(backupDirectory,"backup/");

	//remove quotes from module path

	for(i=0;i<(int)strlen(moduleFileName);++i)
	{
		if(!moduleFileName[i]) break;

		if(moduleFileName[i]=='\"') strcpy(&moduleFileName[i],&moduleFileName[i+1]);
	}

	//load config

	strcpy(configFileName,baseDirectory);
	strcat(configFileName,CONFIG_NAME);

	config_open(configFileName);

	strcpy(textScreenFont, config_read_string("TextScreenFont", "vga.bmp"));
	strcpy(textScreenColor, config_read_string("TextScreenColor", "vga.bmp"));
	strcpy(engineFileName, config_read_string("DefaultEngine", "test.1te"));

	textScreenWidth  =config_read_int("TextScreenWidth",64);
	textScreenHeight =config_read_int("TextScreenHeight",24);
	curPageStep      =config_read_int("PageUpDownStep",16);
	curPageLargeStep =config_read_int("PageUpDownLargeStep",64);
	soundBufferLength=config_read_int("SDLSoundBufferSize",4096);
	logEnable        =config_read_int("EnableLog",1)?true:false;
	backupPeriod     =config_read_int("BackupTime",0);
	asJITEnable      =config_read_int("ASJITEnable",1);

	config_close();

	if(logEnable)
	{
		strcpy(logFileName,baseDirectory);
		strcat(logFileName,"1tracker.log");

		log_open(logFileName);
	}

	if(textScreenWidth <64)  textScreenWidth =64;
	if(textScreenWidth >128) textScreenWidth =128;
	if(textScreenHeight<16)  textScreenHeight=16;
	if(textScreenHeight>128) textScreenHeight=128;

	if(soundBufferLength<1024)  soundBufferLength=1024;
	if(soundBufferLength>65536) soundBufferLength=65536;

	emu_speed_factor = 1;

	//init SDL

	if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO)<0)
	{
		log_add("Error: Can't initialize SDL\n");
		exit(-1);
	}

	atexit(cleanup);

	//load font

	strcpy(fontFileName,baseDirectory);
	strcat(fontFileName, "fonts/");
	strcat(fontFileName, textScreenFont);

	if(!load_font(fontFileName))
	{
		log_add("Error: Can't load font '%s'\n",fontFileName);
		exit(-1);
	}

	//load color scheme
	
	strcpy(colorFileName, baseDirectory);
	strcat(colorFileName, "colors/");
	strcat(colorFileName, textScreenColor);

	if (!load_color(colorFileName))
	{
		log_add("Error: Can't load color scheme '%s'\n", colorFileName);
		exit(-1);
	}
	
	//initialize video

	fullScreen = false;

	screenWidth = charWidth *textScreenWidth;
	screenHeight = charHeight*textScreenHeight;

	sdl_set_video(false);

	clear_screen(COL_PROGRAM_BACKGROUND);
	update_screen();

	//initialize audio

	soundRenderEnable=false;

	sndReq=(SDL_AudioSpec*)malloc(sizeof(SDL_AudioSpec));
	sndObt=(SDL_AudioSpec*)malloc(sizeof(SDL_AudioSpec));
	sndReq->freq=44100;
	sndReq->format=AUDIO_S16SYS;
	sndReq->samples=soundBufferLength;
	sndReq->channels=2;
	sndReq->callback=sdl_sound_callback;
	sndReq->userdata=NULL;
	
	if(SDL_OpenAudio(sndReq,sndObt)<0) critical_error("[SDL] Unable to set audio: %s\n",SDL_GetError());

	if(sndReq->freq    !=sndObt->freq
     ||sndReq->format  !=sndObt->format
	 ||sndReq->channels!=sndObt->channels) critical_error("[SDL] Unable to get requested audio buffer (44100hz 16bit stereo)\n");

	free(sndReq);

	sndReq=sndObt;

	SDL_PauseAudio(0);

	//init AngelScript

	if(!script_init()) critical_error("[AS] Can't initialize AngelScript\n");

#ifndef _DEBUG
	check_engine_updates();
#endif

	//main loop

	editStyle = EDIT_STYLE_LINEAR;

	set_edit_mode(MODE_EDIT_SONG);

	memset(copyPasteData,0,sizeof(copyPasteData));
	memset(copyPasteType,0,sizeof(copyPasteType));

	altMenu = false;
	quietMode = false;

	curAutoStep=1;
	curOctave=4;
	curEnterHold=false;
	curShiftHold=false;
	curResetValue=true;

	copyPasteWidth=0;
	copyPasteHeight=0;

	module_clear();

	if(strlen(moduleFileName)) module_load(moduleFileName);

	backupPrevTime=SDL_GetTicks();

	while(SDL_WaitEvent(&event)>=0)
	{
		switch(event.type)
		{
		case SDL_KEYDOWN:
			{
				switch(event.key.keysym.sym)
				{
				case SDLK_ESCAPE:
					{
						strcpy(engineErrorMessage, "");
					}
					break;

				case SDLK_F10:
					{
						if (altMenu&&editMode == MODE_EDIT_SONG) exit(0);
					}
					break;

				case SDLK_RETURN:
					{
						if(event.key.keysym.mod&(KMOD_RALT|KMOD_LALT))
						{
							fullScreen^=true;
							sdl_set_video(true);
						}
					}
					break;
				}
			}
			break;

		case SDL_USEREVENT:
			{
				switch(event.user.code)
				{
				case EVENT_LOAD_SONG:
					{
						if(strlen((char*)event.user.data1))
						{
							module_load((char*)event.user.data1);
							backupPrevTime=SDL_GetTicks();
						}
					}
					break;

				case EVENT_SAVE_SONG:
					{
						if(strlen((char*)event.user.data1))
						{
							module_save((char*)event.user.data1);
							backupPrevTime=SDL_GetTicks();
						}
					}
					break;

				case EVENT_LOAD_INSTRUMENT:
					{
						if(strlen((char*)event.user.data1))
						{
							instrument_load((char*)event.user.data1,curInstrument);
						}
					}
					break;

				case EVENT_SCRIPT_FILE_SELECT:
				{
					if (strlen((char*)event.user.data1))
					{
						asIScriptContext* context = script_context(engineFileName, "void SelectFileCallback(string)");

						if (!context) critical_error("[AS] Error during song compilation, details are in the log\n");

						string filename = (char*)event.user.data1;

						context->SetArgObject(0, &filename);

						script_execute(context);
					}
				}
				break;

				case EVENT_SAVE_INSTRUMENT:
					{
						if(strlen((char*)event.user.data1))
						{
							instrument_save((char*)event.user.data1,curInstrument);
						}
					}
					break;

				case EVENT_EXPORT_FILE:
					{
						if(strlen((char*)event.user.data1))
						{
							file=fopen((char*)event.user.data1,"wb");

							if(file&&scriptGeneratedData&&scriptGeneratedDataSize)
							{
								fwrite(scriptGeneratedData,scriptGeneratedDataSize,1,file);
								fclose(file);
							}
						}
					}
					break;
				}
			}
			break;

		case SDL_QUIT:
			do_backup();
			exit(0);
			break;
		}

		if (editMode != MODE_INPUT) clear_screen(COL_PROGRAM_BACKGROUND);

		switch(editMode)
		{
		case MODE_EDIT_SONG:
			{
				editor_events(event);
				draw_pattern();
				draw_marker_list();
				draw_settings();
			}
			break;

		case MODE_INPUT:
			{
				input_events(event);
				draw_input();
			}
			break;

		case MODE_SELECT_FILE:
		case MODE_SELECT_FILE_INPUT_DONE:
		case MODE_SELECT_FILE_OVERWRITE:
			{
				select_file_events(event);
				draw_select_file();
			}
			break;

		case MODE_INSTRUMENT_EDITOR:
		case MODE_INSTRUMENT_INPUT_DONE:
			{
				instrument_editor_events(event);
				draw_instrument_editor();
			}
			break;

		case MODE_INFO:
			{
				info_events(event);
				draw_info();
			}
			break;

		case MODE_SELECT_ENGINE:
			{
				select_engine_events(event);
				draw_select_engine();
			}
			break;

		case MODE_HELP:
			{
				help_events(event);
				draw_help();
			}
			break;

		case MODE_EXPORT:
			{
				export_events(event);
				draw_export();
			}
			break;
		}

		update_screen();

		if(backupPeriod)
		{
			if((SDL_GetTicks()-backupPrevTime)>=(backupPeriod*1000*60))
			{
#ifdef WIN32
				_mkdir(backupDirectory);
#else
				mkdir(backupDirectory, S_IRWXU);
#endif
				do_backup();

				backupPrevTime=SDL_GetTicks();
			}
		}
	}

	exit(0);
}