//#define PROFILE_SCRIPT



//macro to register script functions

#define AS_REG(f,n) { if(script_engine->RegisterGlobalFunction((n),asFUNCTION((f)),asCALL_CDECL)<0) return false; }


char scriptLoadedFileName[MAXPATH];
string scriptBinaryData;

int scriptProfileTime;



//script functions

void Print(string &msg)
{
	log_add("%s",msg.c_str());
}



void ProfileStart(void)
{
	scriptProfileTime = SDL_GetTicks();
}



void ProfileEnd(void)
{
	log_add("%i ms\n", SDL_GetTicks() - scriptProfileTime);
}



unsigned int Random(void)
{
	return rand();
}



string GetBaseDirectory(void)
{
	string dir;

	dir = baseDirectory;

	return dir;
}



string LoadBinaryFileAsString(string &filename,int offset,int length)
{
	FILE *file;
	unsigned char *tmp;
	int i,size;

	file=fopen(filename.c_str(),"rb");

	scriptBinaryData.resize(0);

	if(file)
	{
		fseek(file,0,SEEK_END);
		size=ftell(file);
		fseek(file,offset,SEEK_SET);

		if(length<0)
		{
			length=size-offset;
		}

		tmp = (unsigned char*)malloc(length);

		fread(tmp,length,1,file);
		fclose(file);

		for(i=0;i<length;++i)
		{
			scriptBinaryData+=('a'+(tmp[i]&15));
			scriptBinaryData+=('a'+(tmp[i]>>4));
		}

		free(tmp);
	}

	return scriptBinaryData;
}



void AddColumn(int channel,int type,int max,int play)
{
	module.columnType    [module.columns]=type;
	module.columnChannel [module.columns]=channel;
	module.columnMaxValue[module.columns]=max;
	module.columnPlay    [module.columns]=play?true:false;

	++module.columns;
}



void AddExportFormat(string &name,string &ext)
{
	add_export_format(name,ext);
}



int note_to_number(const char* note)
{
	int n;

	switch(note[0])
	{
	case 'C': case 'c': n=0; break;
	case 'D': case 'd': n=2; break;
	case 'E': case 'e': n=4; break;
	case 'F': case 'f': n=5; break;
	case 'G': case 'g': n=7; break;
	case 'A': case 'a': n=9; break;
	case 'B': case 'b': n=11; break;
	default:  n=0;
	}

	if(note[1]=='#') ++n;

	n+=((note[2]-'0')*12);

	return n;
}



void SetNoteRange(string &lowest,string &highest)
{
	module.lowestNote =note_to_number(lowest .c_str())+2;
	module.highestNote=note_to_number(highest.c_str())+2;
}



void SetPatternHeader(string &header)
{
	strcpy(engineHeader,header.c_str());
}



void SetInstrumentsNumber(int number)
{
	module.instruments=number;
}



void SetLoopSupport(int enable)
{
	module.loopSupport=enable?true:false;
}



void SetSpeedRange(int min,int max,int set)
{
	module.speed_min = min;
	module.speed_max = max;
	module.speed = set;
}



void SetTitle(string &title)
{
	strcpy(selectEngineTitle,title.c_str());
}



void SetAbout(string &about)
{
	strcpy(selectEngineAbout,about.c_str());
}



void SetExtraInfo(string &info)
{
	infoExtra=info;
}



int GetSongData(int row,int column)
{
	return module.rows[row].column[column];
}



int GetSongLength(void)
{
	return module_find_first_empty_row();
}



int GetSongLoopStart(void)
{
	return module.loopStart;
}



int GetSongLoopEnd(void)
{
	return module.loopEnd;
}



int GetSongSpeed(void)
{
	return module.speed;
}



string GetSongTitle(void)
{
	return string(module.title);
}



string GetSongAuthor(void)
{
	return string(module.author);
}



int GetInstrumentValue(int instrument,int offset)
{
	return module.instrument[instrument].data[offset];
}



void SetInstrumentValue(int instrument,int offset,int value)
{
	module.instrument[instrument].data[offset] = value;
}



int GetChannelMute(int channel)
{
	return channelMute[channel]?1:0;
}



int GetCurrentChannel(void)
{
	return module.columnChannel[curColumn];
}



int GetCurrentInstrument(void)
{
	return curInstrument;
}



void ItemUnsignedByte(string &caption,int min,int max,int byte)
{
	instrument_gui_add_number(INSGUI_UNSIGNED_BYTE,caption.c_str(),min,max,byte);
}



void ItemUnsignedWord(string &caption,int min,int max,int byte)
{
	instrument_gui_add_number(INSGUI_UNSIGNED_WORD,caption.c_str(),min,max,byte);
}



void ItemSignedByte(string &caption,int min,int max,int byte)
{
	instrument_gui_add_number(INSGUI_SIGNED_BYTE,caption.c_str(),min,max,byte);
}



void ItemSignedWord(string &caption,int min,int max,int byte)
{
	instrument_gui_add_number(INSGUI_SIGNED_WORD,caption.c_str(),min,max,byte);
}



void ItemHexString(string &caption,int len,int byte)
{
	instrument_gui_add_hex_string(caption.c_str(), len, 0, 0, byte, false);
}



void ItemOption(string &caption,string &options,int byte)
{
	instrument_gui_add_option(caption.c_str(),options.c_str(),byte);
}



void ItemLabel(string &caption)
{
	instrument_gui_add_label(caption.c_str());
}



void ItemEnvelope(string &caption,int len, int max, int res,int byte)
{
	instrument_gui_add_hex_string(caption.c_str(), len, max, res, byte, true);
}



void ItemSelectFile(string &caption, string &fileext)
{
	instrument_gui_add_select_file(caption.c_str(), fileext.c_str());
}



void CompileError(string &msg)
{
	strcpy(engineErrorMessage,msg.c_str());
}



void DumpOutput(asIScriptGeneric* gen)
{
	CScriptArray *data;
	int i;

	if(scriptGeneratedData) free(scriptGeneratedData);

	data=*(CScriptArray**)gen->GetAddressOfArg(0);
	
	scriptGeneratedDataSize=data->GetSize();
	scriptGeneratedData=(unsigned char*)malloc(scriptGeneratedDataSize);

	for(i=0;i<scriptGeneratedDataSize;++i) scriptGeneratedData[i]=*((unsigned char*)data->At(i));
}



void CheatGMESpeedFactor(int factor)
{
	emu_speed_factor = factor;
}



// Implement a simple message callback function

void script_callback(const asSMessageInfo *msg,void *param)
{
	const char *type;

	if(msg->type==asMSGTYPE_WARNING) type="WARN"; else if(msg->type==asMSGTYPE_INFORMATION) type="INFO"; else type="ERR ";

	log_add("%s (%d, %d) : %s : %s\n",msg->section,msg->row,msg->col,type,msg->message);
}



bool script_engine_start(void)
{
	script_engine = asCreateScriptEngine(ANGELSCRIPT_VERSION);

	if (!script_engine) return false;

	if (asJITEnable)
	{
		script_engine_jit = new asCJITCompiler(0);

		script_engine->SetEngineProperty(asEP_INCLUDE_JIT_INSTRUCTIONS, 1);
		script_engine->SetJITCompiler(script_engine_jit);
	}

	if(script_engine->SetMessageCallback(asFUNCTION(script_callback),0,asCALL_CDECL)<0) return false;

	RegisterStdString(script_engine);
	RegisterScriptArray(script_engine,true);

	script_engine->SetEngineProperty(asEP_COMPILER_WARNINGS      ,1);//emit warnings
	script_engine->SetEngineProperty(asEP_SCRIPT_SCANNER         ,0);//ASCII encoding
	script_engine->SetEngineProperty(asEP_ALLOW_MULTILINE_STRINGS,true);//allow multiline string literals
	script_engine->SetEngineProperty(asEP_USE_CHARACTER_LITERALS ,true);//allow character literals
	script_engine->SetEngineProperty(asEP_BUILD_WITHOUT_LINE_CUES,true);

	AS_REG(Print                 ,"void Print(const string &in)");
	AS_REG(ProfileStart          ,"void ProfileStart(void)");
	AS_REG(ProfileEnd            ,"void ProfileEnd(void)");
	AS_REG(Random                ,"uint Random(void)");
	AS_REG(LoadBinaryFileAsString,"const string LoadBinaryFileAsString(string &in name,int offset,int length)");
	AS_REG(GetBaseDirectory      ,"string GetBaseDirectory(void)");

	AS_REG(AddColumn           ,"void AddColumn(int channel,int type,int max,int play)");
	AS_REG(AddExportFormat     ,"void AddExportFormat(string &in name,string &in ext)");

	AS_REG(SetNoteRange        ,"void SetNoteRange(string &in lowest,string &in highest)");
	AS_REG(SetPatternHeader    ,"void SetPatternHeader(string &in header)");
	AS_REG(SetInstrumentsNumber,"void SetInstrumentsNumber(int number)");
	AS_REG(SetLoopSupport      ,"void SetLoopSupport(int enable)");
	AS_REG(SetSpeedRange       ,"void SetSpeedRange(int min,int max,int set)");
	AS_REG(SetTitle            ,"void SetTitle(const string &in title)");
	AS_REG(SetAbout            ,"void SetAbout(const string &in about)");
	AS_REG(SetExtraInfo        ,"void SetExtraInfo(const string &in info)");

	AS_REG(GetSongData         ,"uint GetSongData(int row,int column)");
	AS_REG(GetSongLength       ,"uint GetSongLength(void)");
	AS_REG(GetSongLoopStart    ,"uint GetSongLoopStart(void)");
	AS_REG(GetSongLoopEnd      ,"uint GetSongLoopEnd(void)");
	AS_REG(GetSongSpeed        ,"uint GetSongSpeed(void)");
	AS_REG(GetSongTitle        ,"const string GetSongTitle(void)");
	AS_REG(GetSongAuthor       ,"const string GetSongAuthor(void)");

	AS_REG(GetInstrumentValue  ,"uint GetInstrumentValue(int instrument,int offset)");
	AS_REG(SetInstrumentValue  ,"void SetInstrumentValue(int instrument,int offset,int value)");

	AS_REG(GetChannelMute      ,"uint GetChannelMute(int channel)");
	AS_REG(GetCurrentChannel   ,"uint GetCurrentChannel(void)");
	AS_REG(GetCurrentInstrument,"uint GetCurrentInstrument(void)");

	AS_REG(ItemUnsignedByte,"void ItemUnsignedByte(const string &in caption,int min,int max,int byte)");
	AS_REG(ItemUnsignedWord,"void ItemUnsignedWord(const string &in caption,int min,int max,int byte)");
	AS_REG(ItemSignedByte  ,"void ItemSignedByte(const string &in caption,int min,int max,int byte)");
	AS_REG(ItemSignedWord  ,"void ItemSignedWord(const string &in caption,int min,int max,int byte)");
	AS_REG(ItemHexString   ,"void ItemHexString(const string &in caption,int len,int byte)");
	AS_REG(ItemOption      ,"void ItemOption(const string &in caption,const string &in options,int byte)");
	AS_REG(ItemLabel       ,"void ItemLabel(const string &in caption)");
	AS_REG(ItemEnvelope    ,"void ItemEnvelope(const string &in caption,int len,int max,int res,int byte)");
	AS_REG(ItemSelectFile  ,"void ItemSelectFile(const string &in caption,const string &in fileext)");

	AS_REG(CompileError    ,"void CompileError(const string &in msg)");

	AS_REG(CheatGMESpeedFactor,"void CheatGMESpeedFactor(int factor)");

	if(script_engine->RegisterGlobalFunction("void DumpOutput(const uint8[]@+ data)",asFUNCTION(DumpOutput),asCALL_GENERIC)<0) return false;

	return true;
}



void script_engine_stop(void)
{
	if (!script_engine) return;

	script_engine->Release();

	if (asJITEnable) delete script_engine_jit;
}



bool script_init(void)
{
	if(!script_engine_start()) return false;

	strcpy(scriptLoadedFileName,"");
	strcpy(engineErrorMessage,"");

	scriptGeneratedData=NULL;

	return true;
}



void script_shut(void)
{
	script_engine_stop();

	if(scriptGeneratedData)
	{
		free(scriptGeneratedData);
		scriptGeneratedData=NULL;
	}
}



asIScriptContext* script_context(char* file,char* funcName)
{
	CScriptBuilder builder;
	asIScriptModule *mod;
	asIScriptFunction *func;
	asIScriptContext *context;
	char scriptName[MAXPATH];

	sprintf(scriptName,"%sengines/%s",baseDirectory,file);

	//int t=SDL_GetTicks();

	if(strcmp(scriptLoadedFileName,scriptName))
	{
		script_engine_stop();
		script_engine_start();
		
		if(builder.StartNewModule(script_engine,"EngineScript")<0)
		{
			log_add("Error: Unrecoverable error while starting a new AngelScript module\n");
			return NULL;
		}

		strcpy(scriptLoadedFileName,scriptName);

		if(builder.AddSectionFromFile(scriptName)<0)
		{
			log_add("Error: There are errors in the script (%s)\n",scriptName);
			return NULL;
		}

		if(builder.BuildModule()<0)
		{
			log_add("Error: There are errors in the script (%s)\n",scriptName);
			return NULL;
		}
	}

	mod=script_engine->GetModule("EngineScript");

	func=mod->GetFunctionByDecl(funcName);

	if(!func)
	{
		log_add("Script function '%s' not found\n",funcName);
		return NULL;
	}

	context=script_engine->CreateContext();

	context->Prepare(func);

	//log_add("Restarting script engine and compiling a new script %ims\n",SDL_GetTicks()-t);

	return context;
}



bool script_execute(asIScriptContext *context)
{
	int r;
	bool error;

	if(!context) return false;

#ifdef PROFILE_SCRIPT
	int t=SDL_GetTicks();
#endif

	r=context->Execute();

	error = (r != asEXECUTION_FINISHED&&r == asEXECUTION_EXCEPTION);

	if (error) log_add("An exception '%s' occurred\n", context->GetExceptionString());

	context->Release();

#ifdef PROFILE_SCRIPT
	log_add("Executing script %ims\n",SDL_GetTicks()-t);
#endif

	return true;
}



bool script_run(char* file,char* f)
{
	char funcName[1024];

	sprintf(funcName,"void %s()",f);

	return script_execute(script_context(file,funcName));
}



void change_engine(void)
{
	bool error;

	music_stop();

	module.columns=0;
	
	clear_export_format();

	error=false;

	emu_speed_factor = 1;

	if(!script_run(engineFileName,"Init")) error=true; else
	if(!script_run(engineFileName,"Info")) error=true;

	if(error) critical_error("[AS] Error during engine change, details could be found in the log file\n");

	strcpy(engineTitle,selectEngineTitle);
}



bool engine_get_info(char* filename)
{
	return script_run(filename,"Info");
}



void change_instrument(void)
{
	if(!script_run(engineFileName,"Instrument")) critical_error("[AS] Error during constructing instrument dialog, details are in the log\n");
}



bool compile_song(int format,int start,bool oneshot,bool mute)
{
	asIScriptContext* context;

	scriptGeneratedDataSize=0;

	context=script_context(engineFileName,"void Compile(uint,uint,uint,uint)");

	if(!context) critical_error("Script error during song compilation, details are in the log\n");

	context->SetArgDWord(0,format);
	context->SetArgDWord(1,start);
	context->SetArgDWord(2,oneshot?1:0);
	context->SetArgDWord(3,mute?1:0);

	script_execute(context);

	if(!scriptGeneratedDataSize) strcpy(engineErrorMessage,"Engine not provided any data");

	return strlen(engineErrorMessage)?false:true;
}