void set_undo(void)
{
	memcpy(&moduleUndo,&module,sizeof(module));

	undoAvailable=true;
}



void undo(void)
{
	unsigned char *src,*dst,temp;
	int i;

	if(undoAvailable)
	{
		src=(unsigned char*)&module;
		dst=(unsigned char*)&moduleUndo;

		for(i=0;i<sizeof(module);++i)
		{
			temp=*src;
			*src++=*dst;
			*dst++=temp;
		}
	}
}



void move_row_cursor(int off)
{
	curRow+=off;

	if(curRow<0) curRow=0;
	if(curRow>=MAX_ROWS) curRow=MAX_ROWS-1;

	curAutoStepCount=0;
	curResetValue=true;
}



void move_row_cursor_prev_marker(void)
{
	int i;

	module.marker[0]|=2;
	module.marker[module.loopStart]|=2;
	module.marker[module.loopEnd  ]|=2;

	for(i=curRow-1;i>=0;--i)
	{
		if(module.marker[i])
		{
			curRow=i;
			break;
		}
	}

	module.marker[0]&=~2;
	module.marker[module.loopStart]&=~2;
	module.marker[module.loopEnd  ]&=~2;
}



void move_row_cursor_next_marker(void)
{
	int i;

	module.marker[MAX_ROWS-1]|=2;
	module.marker[module.loopStart]|=2;
	module.marker[module.loopEnd  ]|=2;

	for(i=curRow+1;i<MAX_ROWS;++i)
	{
		if(module.marker[i])
		{
			curRow=i;
			break;
		}
	}

	module.marker[MAX_ROWS-1]&=~2;
	module.marker[module.loopStart]&=~2;
	module.marker[module.loopEnd  ]&=~2;
}



void delete_pos_column(int row,int column)
{
	int i;

	if(row>=MAX_ROWS) return;

	if(column<0)
	{
		for (i = row - 1; i < MAX_ROWS - 1; ++i)
		{
			module.marker[i] = module.marker[i + 1];
			strcpy(module.markerName[i], module.markerName[i + 1]);
		}

		module.marker[MAX_ROWS - 1] = 0;
		strcpy(module.markerName[MAX_ROWS - 1], "");
	}
	else
	{
		for(i=row-1;i<MAX_ROWS-1;++i) module.rows[i].column[column]=module.rows[i+1].column[column];

		module.rows[MAX_ROWS-1].column[column]=0;
	}
}



void insert_pos_column(int row,int column)
{
	int i;

	if(row>=MAX_ROWS) return;

	if(column<0)
	{
		for (i = MAX_ROWS - 1; i > row; --i)
		{
			module.marker[i] = module.marker[i - 1];
			strcpy(module.markerName[i], module.markerName[i - 1]);
		}

		module.marker[row] = 0;
		strcpy(module.markerName[row], "");
	}
	else
	{
		for(i=MAX_ROWS-1;i>row;--i) module.rows[i].column[column]=module.rows[i-1].column[column];

		module.rows[row].column[column]=0;
	}
}



void update_selection(void)
{
	if(curShiftHold)
	{
		curSelectionXE=curColumn;
		curSelectionYE=curRow;

		if(curSelectionXS<curSelectionXE)
		{
			curSelectionX=curSelectionXS;
			curSelectionWidth=curSelectionXE-curSelectionXS+1;
		}
		else
		{
			curSelectionX=curSelectionXE;
			curSelectionWidth=curSelectionXS-curSelectionXE+1;
		}

		if(curSelectionYS<curSelectionYE)
		{
			curSelectionY=curSelectionYS;
			curSelectionHeight=curSelectionYE-curSelectionYS+1;
		}
		else
		{
			curSelectionY=curSelectionYE;
			curSelectionHeight=curSelectionYS-curSelectionYE+1;
		}
	}

	if(!curShiftHold||!curSelectionActive)
	{
		curSelectionActive=false;
		curSelectionXS=-1;
		curSelectionYS=-1;
		curSelectionXE=-1;
		curSelectionYE=-1;
		curSelectionWidth=0;
		curSelectionHeight=0;
	}

	curAutoStepCount=0;
	curResetValue=true;
}



void set_selection(int x,int y,int wdt,int hgt)
{
	curSelectionX = x;
	curSelectionY = y;
	curSelectionXS=x;
	curSelectionYS=y;
	curSelectionXE=x+wdt-1;
	curSelectionYE=y+hgt-1;
	curSelectionWidth=wdt;
	curSelectionHeight=hgt;

	curSelectionActive = true;
}



void select_channel(int col)
{
	int i,x,y,w,h,newh,channel,cmin,cmax;

	cmin = curRow;
	cmax = curRow;

	while (cmin > 0)
	{
		if (module.marker[cmin] || (module.loopEnd < (MAX_ROWS - 1) && cmin == module.loopEnd)) break;

		--cmin;
	}

	while (cmax < MAX_ROWS)
	{
		++cmax; 
		
		if (module.marker[cmax] || (module.loopEnd < (MAX_ROWS - 1) && cmax == module.loopEnd)) break;
	}

	if (cmin == 0 && cmax == MAX_ROWS&&curSelectionHeight == MAX_ROWS)
	{
		cmin = 0;
		cmax = 0;
	}

	newh = cmax - cmin;

	if (!curSelectionActive || (curSelectionY != cmin&&curSelectionHeight != newh) || curSelectionHeight==MAX_ROWS)
	{		
		y = cmin;
		h = newh;
	}
	else
	{
		y = 0;
		h = MAX_ROWS;
	}

	if (col >= 0)
	{
		channel = module.columnChannel[col];

		if (channel < 0)
		{
			x = curColumn;
			w = 1;
		}
		else
		{
			cmin = module.columns;
			cmax = 0;

			for (i = 0; i < module.columns; ++i)
			{
				if (module.columnChannel[i] == channel)
				{
					if (i < cmin) cmin = i;
					if (i > cmax) cmax = i;
				}
			}

			x = cmin;
			w = cmax - cmin + 1;
		}
	}
	else
	{
		x = 0;
		w = module.columns;
	}

	set_selection(x, y, w, h);
}



void selection_copy_cut(bool copy,bool cut,bool shift)
{
	int i,j;

	for(i=0;i<curSelectionWidth;++i)
	{
		copyPasteType[i]=module.columnType[curSelectionX+i];

		if(copy) for(j=0;j<curSelectionHeight;++j) copyPasteData[j].column[i]=module.rows[curSelectionY+j].column[curSelectionX+i];

		if(cut)
		{
			for(j=0;j<curSelectionHeight;++j) module.rows[curSelectionY+j].column[curSelectionX+i]=0;

			if(shift)
			{
				for(j=curSelectionY;j<MAX_ROWS;++j) module.rows[j].column[curSelectionX+i]=module.rows[j+curSelectionHeight].column[curSelectionX+i];

				for(j=MAX_ROWS-curSelectionHeight;j<MAX_ROWS;++j) module.rows[j].column[curSelectionX+i]=0;
			}
		}
	}

	if(copy)
	{
		copyPasteWidth =curSelectionWidth;
		copyPasteHeight=curSelectionHeight;
	}

	if(cut&&shift)
	{
		if(curSelectionWidth==module.columns)
		{
			for (j = curSelectionY; j < MAX_ROWS; ++j)
			{
				module.marker[j] = module.marker[j + curSelectionHeight];
				strcpy(module.markerName[j], module.markerName[j + curSelectionHeight]);
			}

			for (j = MAX_ROWS - curSelectionHeight; j < MAX_ROWS; ++j)
			{
				module.marker[j] = 0;
				strcpy(module.markerName[j], "");
			}
		}

		curRow=curSelectionY;
	}

	curSelectionActive=false;

	update_selection();
}



void selection_paste(bool shift)
{
	int i,j,value,type;

	for(i=0;i<copyPasteWidth;++i)
	{
		if(shift)
		{
			for(j=MAX_ROWS-1-copyPasteHeight;j>=curRow;--j) module.rows[j+copyPasteHeight].column[curColumn+i]=module.rows[j].column[curColumn+i];
		}

		type=copyPasteType[i];

		if(module.columnType[curColumn+i]==type)
		{
			for(j=0;j<copyPasteHeight;++j)
			{
				value=copyPasteData[j].column[i];

				switch(type)
				{
				case COLUMN_TYPE_NOTE:
					{
						if(value>1)
						{
							if(value<module.lowestNote)  value=module.lowestNote;
							if(value>module.highestNote) value=module.highestNote;
						}
					}
					break;

				default:
					{
						if(value>module.columnMaxValue[curColumn+i]) value=module.columnMaxValue[curColumn+i];
					}
				}

				module.rows[curRow+j].column[curColumn+i]=value;
			}
		}
	}

	if(shift&&copyPasteWidth==module.columns)
	{
		for (j = MAX_ROWS - 1 - copyPasteHeight; j >= curRow; --j)
		{
			module.marker[j + copyPasteHeight] = module.marker[j];
			strcpy(module.markerName[j + copyPasteHeight], module.markerName[j]);
		}
	}

	curSelectionActive=false;

	update_selection();
}



void selection_expand(void)
{
	int i,j,row;

	if(curSelectionHeight<2) return;

	set_undo();

	row=curSelectionY+1;

	for(i=0;i<curSelectionHeight;++i)
	{
		for (j = 0; j<curSelectionWidth; ++j) insert_pos_column(row, curSelectionX + j);

		insert_pos_column(row, -1);

		row += 2;
	}

	curSelectionActive=false;

	update_selection();
}



void selection_shrink(void)
{
	int i, j, row;

	if(curSelectionHeight<2) return;

	set_undo();

	row=curSelectionY+2;

	for(i=0;i<(curSelectionHeight>>1);++i)
	{
		for (j = 0; j<curSelectionWidth; ++j) delete_pos_column(row, curSelectionX + j);

		delete_pos_column(row, -1);

		++row;
	}

	curSelectionActive=false;

	update_selection();
}



void selection_transpose(int trans)
{
	int i,j,note;

	for(i=0;i<curSelectionHeight;++i)
	{
		for(j=0;j<curSelectionWidth;++j)
		{
			if(module.columnType[curSelectionX+j]==COLUMN_TYPE_NOTE)
			{
				note=module.rows[curSelectionY+i].column[curSelectionX+j];

				if(note>1)
				{
					note+=trans;

					if(note<module.lowestNote)  note=module.lowestNote;
					if(note>module.highestNote) note=module.highestNote;

					module.rows[curSelectionY+i].column[curSelectionX+j]=note;
				}
			}
		}
	}
}



void selection_replicate(void)
{
	int i,j;

	for(i=0;i<curSelectionWidth;++i)
	{
		for(j=MAX_ROWS-1;j>=curSelectionY+curSelectionHeight;--j)
		{
			module.rows[j].column[curSelectionX+i]=module.rows[j-curSelectionHeight].column[curSelectionX+i];
		}
	}

	if(curSelectionWidth==module.columns)
	{
		for (j = MAX_ROWS - 1; j >= curSelectionY + curSelectionHeight; --j)
		{
			module.marker[j] = module.marker[j - curSelectionHeight];
			strcpy(module.markerName[j], module.markerName[j - curSelectionHeight]);
		}
	}

	curSelectionY +=curSelectionHeight;
	curSelectionYS+=curSelectionHeight;
	curSelectionYE+=curSelectionHeight;

	if(curSelectionY >=MAX_ROWS) curSelectionY =MAX_ROWS-1;
	if(curSelectionYS>=MAX_ROWS) curSelectionYS=MAX_ROWS-1;
	if(curSelectionYE> MAX_ROWS) curSelectionYE=MAX_ROWS;

	curRow+=curSelectionHeight;

	if(curRow>=MAX_ROWS) curRow=MAX_ROWS-1;
}



void change_channel(int dir)
{
	int i;

	i=module.columnChannel[curColumn];

	if(dir<0)
	{
		while(i==module.columnChannel[curColumn]||module.columnType[curColumn]==COLUMN_TYPE_EMPTY)
		{
			--curColumn;

			if(curColumn<0) curColumn=module.columns-1;
		}
	}
	else
	{
		while(i==module.columnChannel[curColumn]||module.columnType[curColumn]==COLUMN_TYPE_EMPTY)
		{
			++curColumn;

			if(curColumn>=module.columns) curColumn=0;
		}
	}
}



int get_cursor_row_y(void)
{
	int y, row, row2x;

	row2x = get_pattern_top_row_2x();
	y = PATTERN_YPOS;

	while (y < (PATTERN_YPOS + PATTERN_HEIGHT))
	{
		row = row2x / 2;

		if (row >= MAX_ROWS) break;

		if (!(row2x & 1))
		{
			if (row && (module.marker[row] || (module.loopEnd < MAX_ROWS - 1 && row == module.loopEnd)))
			{
				++y;
			}
		}
		else
		{
			if (row == curRow) break;

			++y;
		}

		++row2x;
	}

	return y;
}



int get_pattern_width(void)
{
	int w, ct, col;

	w = 0;

	for (col = 0; col < module.columns; ++col)
	{
		ct = module.columnType[col];

		switch (ct)
		{
		case COLUMN_TYPE_EMPTY: ++w; break;
		case COLUMN_TYPE_NOTE: w += 3; break;
		default: w += module.columnType[col];
		}
	}

	return w;
}



void editor_events(SDL_Event event)
{
	int i, key, ct, note, channel, row;
	bool shift,ctrl;

	shift=event.key.keysym.mod&(KMOD_LSHIFT|KMOD_RSHIFT)?true:false;
	ctrl= event.key.keysym.mod&(KMOD_LCTRL |KMOD_RCTRL )?true:false;

	key = event.key.keysym.sym;

	if (event.type == SDL_KEYDOWN)
	{
		if (key == SDLK_LCTRL || key == SDLK_RCTRL) altMenu = true;
	}
	if (event.type == SDL_KEYUP)
	{
		if (key == SDLK_LCTRL || key == SDLK_RCTRL) altMenu = false;
	}

	switch(event.type)
	{
	case SDL_KEYDOWN:
		{
			ct=module.columnType[curColumn];

			if (!altMenu)
			{
				switch (key)
				{
					case SDLK_F1:
					{
						set_edit_mode(MODE_HELP);
					}
					break;

					case SDLK_F2:
					{
						set_select_file("Save song file", songDirectory, MODE_EDIT_SONG, SONG_EXTENSION, true, EVENT_SAVE_SONG, selectFileSongLast);
					}
					break;

					case SDLK_F3:
					{
						set_select_file("Load song file", songDirectory, MODE_EDIT_SONG, SONG_EXTENSION, true, EVENT_LOAD_SONG, selectFileSongLast);
					}
					break;

					case SDLK_F4:
					{
						set_instrument_editor();
					}
					break;

					case SDLK_F5:
					{
						music_toggle(false);
					}
					break;

					case SDLK_F6:
					{
						set_info();
					}
					break;

					case SDLK_F7:
					{
						set_select_engine();
					}
					break;

					case SDLK_F8:
					{
						set_export();
					}
					break;
				}
			}
			else
			{
				switch (key)
				{
					case SDLK_F1:
					{
						if (editStyle == EDIT_STYLE_CHOPPED) editStyle = EDIT_STYLE_LINEAR; else editStyle = EDIT_STYLE_CHOPPED;
					}
					break;

					case SDLK_F2:
					{
						if (!quietMode) quietMode = true; else quietMode = false;
					}
					break;

					case SDLK_F3:
					{
						do_backup();
						module_clear();
					}
					break;
				}
			}

			if(!ctrl)
			{
				switch(key)
				{
				case SDLK_UP://move cursor up one row
					{
						move_row_cursor(-1);
						update_selection();
					}
					break;

				case SDLK_DOWN://move cursor down one row
					{
						move_row_cursor(1);
						update_selection();
					}
					break;

				case SDLK_PAGEUP://move cursor page up
					{
						move_row_cursor(-curPageStep);
						update_selection();
					}
					break;

				case SDLK_PAGEDOWN://move cursor page down
					{
						move_row_cursor(curPageStep);
						update_selection();
					}
					break;

				case SDLK_HOME://move cursor to a previous marker
					{
						move_row_cursor_prev_marker();
						update_selection();
					}
					break;

				case SDLK_END://move cursor to a next marker
					{
						move_row_cursor_next_marker();
						update_selection();
					}
					break;

				case SDLK_LEFT://move cursor left
					{
						--curColumn;

						if(curColumn<0) curColumn=module.columns-1;

						while(module.columnType[curColumn]==COLUMN_TYPE_EMPTY) --curColumn;

						update_selection();
					}
					break;

				case SDLK_RIGHT://move cursor right
					{
						++curColumn;

						if(curColumn>=module.columns) curColumn=0;

						while(module.columnType[curColumn]==COLUMN_TYPE_EMPTY) ++curColumn;

						update_selection();
					}
					break;

				case SDLK_TAB://move cursor to the next or previous channel
					{
						if(shift) change_channel(-1); else change_channel(1);

						update_selection();
					}
					break;

				case SDLK_DELETE://clear a field under cursor or a selection
					{
						set_undo();

						if(curSelectionWidth&&curSelectionHeight)
						{
							selection_copy_cut(false,true,false);
						}
						else
						{
							module.rows[curRow].column[curColumn]=0;
							move_row_cursor(curAutoStep);
							update_selection();
						}
					}
					break;

				case SDLK_BACKQUOTE://increment or decrement autostep
					{
						if(shift)
						{
							if(curAutoStep) --curAutoStep;
						}
						else
						{
							if(curAutoStep<99) ++curAutoStep;
						}
					}
					break;

				case SDLK_KP_MULTIPLY://increment octave
					{
						if(curOctave<9) ++curOctave;
					}
					break;

				case SDLK_KP_MINUS://decrement octave
					{
						--module.speed;

						if (module.speed < module.speed_min) module.speed = module.speed_min;
					}
					break;

				case SDLK_KP_PLUS://increase global speed
					{
						++module.speed;

						if (module.speed > module.speed_max) module.speed = module.speed_max;
					}
					break;

				case SDLK_KP_DIVIDE://decrease global speed
					{
						if(curOctave) --curOctave;
					}
					break;

				case SDLK_LSHIFT://start selection
				case SDLK_RSHIFT:
					{
						curShiftHold=true;

						if(!curSelectionActive)
						{
							curSelectionXS=curColumn;
							curSelectionYS=curRow;
							curSelectionActive=true;
						}

						update_selection();
					}
					break;

				case SDLK_SPACE://toggle a marker
					{
						module.marker[curRow]^=1;
					}
					break;

				case SDLK_INSERT://insert an empty position into current channel with shift
					{
						set_undo();

						channel=module.columnChannel[curColumn];

						for(i=0;i<MAX_COLUMNS;++i)
						{
							if(module.columnChannel[i]==channel) insert_pos_column(curRow,i);
						}
					}
					break;

				case SDLK_BACKSPACE://delete position of current channel with shift
					{
						set_undo();

						channel=module.columnChannel[curColumn];

						for(i=0;i<MAX_COLUMNS;++i)
						{
							if(module.columnChannel[i]==channel) delete_pos_column(curRow,i);
						}

						move_row_cursor(-1);
						update_selection();
					}
					break;

				case SDLK_8:
					{
						if(!shift) music_play_row();
					}
					break;

				case SDLK_RETURN:
					{
						if(!curEnterHold)
						{
							curEnterHold=true;
							music_play_from_cursor();
						}
					}
					break;
				}

				if(!ctrl&&!shift)
				{
					switch(ct)
					{
					case COLUMN_TYPE_NOTE://enter note
						{
							for(i=0;i<noteKeysCount;i+=2)
							{
								if(noteKeys[i]==key)
								{
									set_undo();

									note=noteKeys[i+1];

									if(note<0)
									{
										note=1;
									}
									else
									{
										note=2+note+curOctave*12;

										if(note<module.lowestNote)  note=module.lowestNote;
										if(note>module.highestNote) note=module.highestNote;
									}

									module.rows[curRow].column[curColumn]=note;

									if (note != 1 && module.columnPlay[curColumn]) if (!quietMode) music_play_row();

									move_row_cursor(curAutoStep);

									break;
								}
							}
						}
						break;

					default://enter numeric value
						{
							if(key>=SDLK_0&&key<=SDLK_9)
							{
								set_undo();

								if(curResetValue)
								{
									i=0;
									curResetValue=false;
								}
								else
								{
									i=module.rows[curRow].column[curColumn];
								}

								i=(i*10+(key-SDLK_0))%(int)pow(10.0f,ct);

								if(i>module.columnMaxValue[curColumn]) i=module.columnMaxValue[curColumn];

								module.rows[curRow].column[curColumn]=i;

								if (i&&module.columnPlay[curColumn]) if (!quietMode) music_play_row();

								++curAutoStepCount;

								if(curAutoStepCount>=module.columnType[curColumn]) move_row_cursor(curAutoStep);
							}
						}
					}

					if(key>=SDLK_KP_0&&key<=SDLK_KP_9)//select octave
					{
						curOctave=key-SDLK_KP_0;
					}
				}
			}

			if(ctrl)
			{
				switch(key)
				{
				case SDLK_PAGEUP:////move cursor large page up
					{
						move_row_cursor(-curPageLargeStep);
						update_selection();
					}
					break;

				case SDLK_PAGEDOWN://move cursor large page down
					{
						move_row_cursor(curPageLargeStep);
						update_selection();
					}

				case SDLK_HOME://set loop start
					{
						if(module.loopSupport)
						{
							set_undo();
							module.loopStart=curRow;
						}
					}
					break;

				case SDLK_END://set loop end
					{
						if(module.loopSupport)
						{
							set_undo();
							module.loopEnd=curRow;
						}
					}
					break;

				case SDLK_a://select all channels
					{
						select_channel(-1);
					}
					break;

				case SDLK_l://select current channel
					{
						select_channel(curColumn);
					}
					break;

				case SDLK_f://name a block
				{
					row = curRow;

					while (1)
					{
						if (module.marker[row]) break;
						if (!row) break;

						--row;
					}

					altMenu = false;

					set_input_string(PATTERN_XPOS + 8 + get_pattern_width(), get_cursor_row_y(), MAX_MARKER_NAME + 1, MODE_EDIT_SONG, true, module.markerName[row]);
				}
				break;

				case SDLK_g://go to line
					{
						set_input_number(PATTERN_XPOS + 1, get_cursor_row_y(), 4, MODE_EDIT_SONG, true, &curRow, false);
					}
					break;

				case SDLK_z://undo
					{
						undo();
					}
					break;

				case SDLK_x://cut selection
					{
						set_undo();
						selection_copy_cut(true,true,false);
					}
					break;

				case SDLK_c://copy selection
					{
						selection_copy_cut(true,false,false);
					}
					break;

				case SDLK_v://paste selection
					{
						set_undo();
						selection_paste(false);
					}
					break;

				case SDLK_0://unmute all
					{
						for(i=0;i<MAX_COLUMNS;++i) channelMute[i]=false;
					}
					break;

				case SDLK_BACKQUOTE://mute all but current channel
					{
						for(i=0;i<MAX_COLUMNS;++i)
						{
							channelMute[i]=(i!=module.columnChannel[curColumn])?true:false;
						}
					}
					break;

				case SDLK_INSERT://insert an empty row with shift
					{
						set_undo();

						for(i=-1;i<module.columns;++i) insert_pos_column(curRow,i);
					}
					break;

				case SDLK_BACKSPACE://delete a row with shift
					{
						set_undo();

						for(i=-1;i<module.columns;++i) delete_pos_column(curRow,i);

						move_row_cursor(-1);
						update_selection();
					}
					break;

				case SDLK_e://expand selection
					{
						selection_expand();
					}
					break;

				case SDLK_s://shrink selection
					{
						selection_shrink();
					}
					break;

				case SDLK_u://transpose semitone up
					{
						set_undo();
						selection_transpose(1);
					}
					break;

				case SDLK_d://transpose semitone down
					{
						set_undo();
						selection_transpose(-1);
					}
					break;

				case SDLK_o://transpose octave up
					{
						set_undo();
						selection_transpose(12);
					}
					break;

				case SDLK_k://transpose octave up
					{
						set_undo();
						selection_transpose(-12);
					}
					break;

				case SDLK_r:
					{
						set_undo();
						selection_replicate();
					}
					break;

				case SDLK_LEFTBRACKET://decrement measure
					{
						if(module.measure>2) --module.measure;
					}
					break;

				case SDLK_RIGHTBRACKET://increment measure
					{
						if(module.measure<1024) ++module.measure;
					}
					break;

				case SDLK_F5://play from row
					{
						music_toggle(true);
					}
					break;

				case SDLK_EQUALS://increase global speed
					{
						if(module.speed<99) ++module.speed;
					}
					break;

				case SDLK_MINUS://decrease global speed
					{
						if(module.speed>1) --module.speed;
					}
					break;
				}

				if (key >= SDLK_1  &&key <= SDLK_9) channelMute[key - SDLK_1] ^= true;//toggle mute for channel 1-9

				if (key >= SDLK_KP_1&&key <= SDLK_KP_9) channelMute[key - SDLK_KP_1] ^= true;
			}

			if(shift&&!ctrl)
			{
				switch(key)
				{
				case SDLK_x://cut selection with shift
					{
						set_undo();
						selection_copy_cut(true,true,true);
					}
					break;

				case SDLK_c://copy selection
					{
						selection_copy_cut(true,false,false);
					}
					break;

				case SDLK_v://paste selection with shift
					{
						set_undo();
						selection_paste(true);
					}
					break;
				}

				if(key>=SDLK_0&&key<=SDLK_9)//select octave
				{
					curOctave=key-SDLK_0;
				}
			}
		}
		break;

	case SDL_KEYUP:
		{
			switch(key)
			{
			case SDLK_LSHIFT:
			case SDLK_RSHIFT:
				{
					curShiftHold=false;
				}
				break;

			case SDLK_RETURN:
				{
					if(curEnterHold)
					{
						curEnterHold=false;
						music_stop();
					}
				}
				break;
			}

			/*for(i=0;i<noteKeysCount;i+=2)
			{
				if(noteKeys[i]==key)
				{
					music_stop();
					break;
				}
			}*/
		}
		break;
	}
}